class BaseOperator:
    operation_template: str = ""


class IsNullOperator(BaseOperator):
    operation_template: str = "{} IS NULL"


class IsNotNullOperator(BaseOperator):
    operation_template: str = "{} IS NOT NULL"


class EqualOperator(BaseOperator):
    operation_template: str = "{} = {}"


class NotEqualOperator(BaseOperator):
    operation_template: str = "{} != {}"


class GreaterOperator(BaseOperator):
    operation_template: str = "{} > {}"


class GreaterEqualOperator(BaseOperator):
    operation_template: str = "{} >= {}"


class LessOperator(BaseOperator):
    operation_template: str = "{} < {}"


class LessEqualOperator(BaseOperator):
    operation_template: str = "{} <= {}"


class InOperator(BaseOperator):
    operation_template: str = "{} IN ({})"


class NotInOperator(BaseOperator):
    operation_template: str = "{} NOT IN ({})"


class LikeOperator(BaseOperator):
    operation_template: str = "{} LIKE {}"


class NotLikeOperator(BaseOperator):
    operation_template: str = "{} NOT LIKE {}"


class ILikeOperator(BaseOperator):
    operation_template: str = "{} NOT ILIKE {}"


class NotILikeOperator(BaseOperator):
    operation_template: str = "{} NOT ILIKE {}"


class BetweenOperator(BaseOperator):
    operation_template: str = "{} BETWEEN {} AND {}"


class ANDOperator(BaseOperator):
    operation_template: str = "AND"


class OROperator(BaseOperator):
    operation_template: str = "OR"


class NotOperator(BaseOperator):
    operation_template: str = "NOT ({})"



class QaspenBaseError(Exception):
    """Base error for all errors."""


class DatabaseUrlParseError(QaspenBaseError):
    """If database from connection url cannot be parsed."""


class FieldBaseError(QaspenBaseError):
    """Base error for all field-related errors."""


class FieldValueValidationError(FieldBaseError):
    """Error for field value validation error."""


class FieldDeclarationError(FieldBaseError):
    """Error for field declaration error."""


class FieldComparisonError(FieldBaseError):
    """Error for error in field comparison.

    Usually if types are incorrect.
    """


class FilterComparisonError(FieldBaseError):
    """Error if there is an mistake in Filter."""


class QueryResultError(QaspenBaseError):
    """Base error for QueryResult exceptions."""


class QueryResultLookupError(QueryResultError):
    """Error for processing result from engine."""



from __future__ import annotations

import dataclasses
import enum
import functools
import operator
from typing import TYPE_CHECKING, Any, Final, Iterable

from qaspen.querystring.querystring import QueryString
from qaspen.statements.combinable_statements.combinations import (
    CombinableExpression,
)
from qaspen.statements.statement import BaseStatement

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.fields.base import Field
    from qaspen.qaspen_types import FieldType
    from qaspen.table.base_table import BaseTable


class Join(CombinableExpression):
    """Main class for PostgreSQL JOINs.

    It can work alone as a class and it's possible
    to inherit from it to create new types of JOINs.

    `join_type` is used for creating QueryString,
    basically it is join type.
    """

    join_type: str = "JOIN"

    def __init__(
        self: Self,
        fields: Iterable[Field[Any]] | None,
        from_table: type[BaseTable],
        join_table: type[BaseTable],
        on: CombinableExpression,
        join_alias: str,
    ) -> None:
        self._from_table: Final[type[BaseTable]] = from_table
        self._join_table: Final[type[BaseTable]] = join_table
        self._based_on: CombinableExpression = on
        self._alias: str = join_alias

        self._fields: list[Field[Any]] | None = None
        if fields:
            self._fields = self._process_select_fields(
                fields=fields,
            )

    def querystring(self: Self) -> QueryString:
        """Build `QueryString`."""
        return QueryString(
            self.join_type,
            self._join_table.schemed_original_table_name(),
            self._join_table._table_meta.alias or self._alias,
            self._based_on.querystring(),
            sql_template="{} {} AS {} ON {}",
        )

    def add_fields(
        self: Self,
        fields: list[Field[Any]],
    ) -> None:
        """Add new fields to the join.

        ### Parameters:
        - `fields`: fields to add.
        """
        processed_fields = self._process_select_fields(
            fields=fields,
        )
        if self._fields:
            self._fields.extend(processed_fields)
            return

        self._fields = processed_fields

    def _prefixed_field(
        self: Self,
        field: Field[FieldType],
    ) -> Field[FieldType]:
        return field._with_prefix(
            prefix=(
                field._field_data.from_table._table_meta.alias or self._alias
            ),
        )

    def _process_select_fields(
        self: Self,
        fields: Iterable[Field[Any]],
    ) -> list[Field[Any]]:
        fields_with_prefix: list[Field[Any]] = [
            self._prefixed_field(field=field) for field in fields
        ]
        return fields_with_prefix

    def _join_fields(self: Self) -> list[Field[Any]] | None:
        return self._fields


class InnerJoin(Join):
    """Class for `INNER JOIN` join type."""

    join_type: str = "INNER JOIN"


class LeftOuterJoin(Join):
    """Class for `LEFT JOIN` join type."""

    join_type: str = "LEFT JOIN"


class RightOuterJoin(Join):
    """Class for `RIGHT JOIN` join type."""

    join_type: str = "RIGHT JOIN"


class FullOuterJoin(Join):
    """Class for `FULL OUTER JOIN` join type."""

    join_type: str = "FULL OUTER JOIN"


class JoinType(enum.Enum):
    """ENUM for JoinType."""

    JOIN: type[Join] = Join
    INNERJOIN: type[InnerJoin] = InnerJoin
    LEFTJOIN: type[LeftOuterJoin] = LeftOuterJoin
    RIGHTJOIN: type[RightOuterJoin] = RightOuterJoin
    FULLOUTERJOIN: type[FullOuterJoin] = FullOuterJoin


@dataclasses.dataclass
class JoinStatement(BaseStatement):
    """Join statement for high-level statements.

    It is used in Select/Update/Insert/Delete Statements.

    `join_expressions` contains all created joins.
    """

    join_expressions: list[Join] = dataclasses.field(
        default_factory=list,
    )

    def join(
        self: Self,
        join_table: type[BaseTable],
        from_table: type[BaseTable],
        on: CombinableExpression,
        join_type: JoinType,
        fields: Iterable[Field[Any]] | None = None,
    ) -> None:
        """Create new join.

        ### Parameters:
        - `join_table`: Table for join.
        - `from_table`: main Table from the query.
        - `on`: `ON` condition (Filter class usually).
        - `join_type`: type of the JOIN.
        - `fields`: fields to select from `join_table`.
        """
        join_alias = (
            join_table._table_meta.alias or join_table.original_table_name()
        )
        self.join_expressions.append(
            join_type.value(
                join_alias=join_alias,
                fields=fields,
                join_table=join_table,
                from_table=from_table,
                on=on,
            ),
        )

    def add_join(
        self: Self,
        *join_expression: Join,
    ) -> None:
        """Add new join.

        ### Parameters:
        - `join_expression`: instance of Join (or its subclasses).
        """
        self.join_expressions.extend(join_expression)

    def querystring(self: Self) -> QueryString:
        """Build `QueryString`."""
        if not self.join_expressions:
            return QueryString.empty()

        final_join: QueryString = functools.reduce(
            operator.add,
            [
                join_expression.querystring()
                for join_expression in self.join_expressions
            ],
        )
        return final_join

    def _retrieve_all_join_fields(
        self: Self,
    ) -> list[Field[Any]]:
        all_joins_fields: list[Field[Any]] = []
        for join_expression in self.join_expressions:
            if join_fields := join_expression._join_fields():
                all_joins_fields.extend(join_fields)
        return all_joins_fields



from __future__ import annotations

from qaspen.querystring.querystring import QueryString


class SQLType:
    """Base class for all PostgreSQL types.

    It has mapping to it's type in python.

    Usually mapped one-to-one to the Field.
    For example, `SmallInt Field` mapped to `SmallInt` type.

    But in some Fields, like `Array`, it's necessary
    to specify additional type.

    Also this and all subclasses are used in pydantic/msgspec
    converting.
    """

    python_type: type | tuple[type, ...]

    @classmethod
    def querystring(cls: type[SQLType]) -> QueryString:
        """Build querystring.

        Usually it equals to upper name of the class.

        ### Returns:
        new `QueryString`.
        """
        return QueryString(
            cls.sql_type(),
            sql_template="{}",
        )

    @classmethod
    def sql_type(cls: type[SQLType]) -> str:
        """Build string Type in PostgreSQL.

        ### Returns:
        string.
        """
        return cls.__name__.upper()



import typing
from urllib.parse import urlparse

from qaspen.exceptions import DatabaseUrlParseError


def parse_database(database_url: str) -> str:
    """Parse database from connection url.

    ### Returns:
    Database name from connection url.
    """
    database_name: typing.Final = urlparse(database_url).path[1:]
    if not database_name:
        parsing_error_message: typing.Final = (
            "Database name cannot be parsed, it's empty"
        )
        raise DatabaseUrlParseError(parsing_error_message)

    return database_name



from __future__ import annotations

import json
from ast import literal_eval
from typing import TYPE_CHECKING, Any, Dict, Final, List, Union

from qaspen.base.operators import AllOperator, AnyOperator
from qaspen.exceptions import FieldDeclarationError, FieldValueValidationError
from qaspen.fields.base import Field
from qaspen.qaspen_types import FieldDefaultType, FieldType
from qaspen.sql_type import complex_types

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.sql_type.base import SQLType


class JsonBase(Field[FieldType]):
    """Base field for JSON and JSONB PostgreSQL fields."""

    def __init__(
        self: Self,
        *args: Any,
        is_null: bool = False,
        db_field_name: str | None = None,
        default: FieldDefaultType[FieldType] = None,
    ) -> None:
        super().__init__(
            *args,
            is_null=is_null,
            default=default,
            db_field_name=db_field_name,
        )

    def _prepare_default_value(
        self: Self,
        default_value: FieldType | None,
    ) -> str:
        """Prepare default value for PostgreSQL DEFAULT statement.

        ### Returns:
        `Any available type for this class`.
        """
        if isinstance(default_value, str):
            try:
                json.loads(default_value)
            except json.decoder.JSONDecodeError as exc:
                validation_err_msg: Final = (
                    f"Default value {default_value} of field "
                    f"{self.__class__.__name__} "
                    f"can't be serialized in PSQL {self._field_type} type.",
                )
                raise FieldValueValidationError(
                    validation_err_msg,
                ) from exc
            return f"'{default_value}'"

        if isinstance(default_value, (dict, list)):
            return self._dump_default(
                default_value=default_value,
            )

        if isinstance(default_value, bytes):
            return self._dump_default(
                literal_eval(
                    default_value.decode("utf-8"),
                ),
            )

        type_err_msg: Final = (
            f"Can't set default value {default_value} for "
            f"{self.__class__.__name__} field",
        )
        raise FieldDeclarationError(type_err_msg)

    def _dump_default(
        self: Self,
        default_value: dict[Any, Any] | list[Any],
    ) -> str:
        dump_value: Final = json.dumps(
            default_value,
            default=str,
        )
        return f"'{dump_value}'"


class JsonField(JsonBase[Union[Dict[Any, Any], str]]):
    """Field for JSON PostgreSQL type."""

    _available_comparison_types: tuple[
        type,
        ...,
    ] = (
        dict,
        list,
        str,
        Field,
        AllOperator,
        AnyOperator,
    )
    _set_available_types: tuple[type, ...] = (dict, list, str)
    _sql_type = complex_types.Json


class JsonbField(JsonBase[Union[Dict[Any, Any], str, bytes]]):
    """Field for JSON PostgreSQL type."""

    _available_comparison_types: tuple[
        type,
        ...,
    ] = (
        bytes,
        dict,
        str,
        list,
        Field,
        AllOperator,
        AnyOperator,
    )
    _set_available_types: tuple[type, ...] = (dict, str, bytes, list)
    _sql_type = complex_types.Jsonb


class ArrayField(Field[List[Any]]):
    """Field for ARRAY PostgreSQL type."""

    _available_comparison_types: tuple[
        type,
        ...,
    ] = (
        list,
        Field,
        AllOperator,
        AnyOperator,
    )
    _set_available_types: tuple[type, ...] = (list,)
    _sql_type = complex_types.Array

    def __init__(
        self: Self,
        *args: Any,
        base_type: type[SQLType],
        is_null: bool = False,
        db_field_name: str | None = None,
        default: FieldDefaultType[list[Any]] = None,
        dimension: int | None = None,
    ) -> None:
        super().__init__(
            *args,
            is_null=is_null,
            default=default,
            db_field_name=db_field_name,
        )

        self.dimension: Final = dimension
        self.base_type: Final = base_type

    def _prepare_default_value(
        self: Self,
        default_value: list[Any] | None,
    ) -> str | None:
        dumped_value = json.dumps(
            default_value,
            default=str,
        )
        return dumped_value.replace("[", "{").replace("]", "}")

    @property
    def _field_type(self: Self) -> str:
        sql_array_type: str = (
            f"{self.base_type.querystring()} {self._sql_type.sql_type()}"
        )
        if self.dimension:
            sql_array_type += f"[{self.dimension}]"

        return sql_array_type



from __future__ import annotations

import sys
from collections import UserDict
from typing import TYPE_CHECKING, Any, Final

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.fields.base import Field


class FieldAlias:
    """Class for field alias.

    It represents single alias.
    """

    def __init__(
        self: Self,
        aliased_field: Field[Any],
    ) -> None:
        """Initialize field alias.

        ### Parameters:
        - `aliased_field`: any Field.
        """
        self.aliased_field: Final = aliased_field


if sys.version_info >= (3, 9):

    class FieldAliases(UserDict[str, FieldAlias]):
        """Class for all aliases."""

        def __init__(self: Self) -> None:
            self.data: dict[str, FieldAlias] = {}

        def add_alias(
            self: Self,
            field: Field[Any],
        ) -> Field[Any]:
            """Add new alias for the field.

            ### Parameters:
            - `field`: any Field.

            ### Returns:
            Field with an alias.
            """
            field_alias: str
            aliased_field: Field[Any]
            if exists_alias := field.alias:
                field_alias = exists_alias
                aliased_field = field
            else:
                aliased_field = field.with_alias(
                    alias_name=field._original_field_name,
                )
                field_alias = aliased_field.alias  # type: ignore[assignment]

            self.data[field_alias] = FieldAlias(
                aliased_field=field,
            )
            return field

        def __str__(self: Self) -> str:
            return str(self.data)

else:

    class FieldAliases(UserDict):  # type: ignore[type-arg]
        """Class for all aliases."""

        def __init__(self: Self) -> None:
            self.data: dict[str, FieldAlias] = {}
            self.last_alias_number: int = 0

        def add_alias(
            self: Self,
            field: Field[Any],
        ) -> Field[Any]:
            """Add new alias for the field.

            ### Parameters:
            - `field`: any Field.

            ### Returns:
            Field with an alias.
            """
            field_alias: str
            aliased_field: Field[Any]
            if exists_alias := field.alias:
                field_alias = exists_alias
                aliased_field = field
            else:
                aliased_field = field.with_alias(
                    alias_name=field._original_field_name,
                )
                field_alias = aliased_field.alias  # type: ignore[assignment]

            self.data[field_alias] = FieldAlias(
                aliased_field=field,
            )
            return field

        def __str__(self: Self) -> str:
            return str(self.data)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from typing_extensions import Self


class QueryString:
    """QueryString for all statements.

    This class is used for building SQL queries.
    All queries must be build with it.

    `add_delimiter` is for `__add__` method.
    """

    add_delimiter: str = " "

    def __init__(
        self: Self,
        *template_arguments: Any,
        sql_template: str,
    ) -> None:
        self.sql_template: str = sql_template
        self.template_arguments: list[Any] = list(template_arguments)

    @classmethod
    def empty(cls: type[QueryString]) -> EmptyQueryString:
        """Create `EmptyQueryString`.

        :returns: EmptyQueryString.
        """
        return EmptyQueryString(sql_template="")

    def querystring(self: Self) -> str:
        """Format QueryString template with arguments.

        ### Returns
        :returns: QueryString as a string.

        Example:
        -------
        ```python
        qs1 = QueryString(
            "good_field",
            "good_table",
            sql_template="SELECT {} FROM {}",
        )
        print(qs1)  # SELECT good_field FROM good_table
        ```
        """
        return self.sql_template.format(
            *self.template_arguments,
        )

    def __add__(
        self: Self,
        additional_querystring: QueryString,
    ) -> Self:
        """Combine two QueryStrings.

        ### Parameters
        :param `additional_querystring`: second QueryString.

        ### Returns
        :returns: self.

        Example:
        -------
        ```python
        qs1 = QueryString(
            "good_field",
            "good_table",
            sql_template="SELECT {} FROM {}",
        )
        qs2 = QueryString(
            "good_field",
            sql_template="ORDER BY {}",
        )
        result_qs = qs1 + qs2
        print(result_qs)
        # SELECT good_field FROM good_table ORDER BY good_field
        ```
        """
        if isinstance(additional_querystring, EmptyQueryString):
            return self

        self.sql_template += (
            f"{self.add_delimiter}{additional_querystring.sql_template}"
        )
        self.template_arguments.extend(
            additional_querystring.template_arguments,
        )
        return self

    def __str__(self: Self) -> str:
        """Return `QueryString` as a string.

        ### Returns
        :returns: string
        """
        return self.querystring()


class EmptyQueryString(QueryString):
    """QueryString without data inside."""

    add_delimiter: str = ""


class OrderByQueryString(QueryString):
    """QueryString for OrderBy clauses."""

    add_delimiter: str = ", "


class FilterQueryString(QueryString):
    """QueryString for FilterStatements like `WHERE`."""

    add_delimiter: str = " AND "


class FullStatementQueryString(QueryString):
    """QueryString for full statements."""

    add_delimiter: str = "; "



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Any

from qaspen.base.sql_base import SQLSelectable
from qaspen.querystring.querystring import QueryString
from qaspen.utils.fields_utils import transform_value_to_sql

if TYPE_CHECKING:
    from typing_extensions import Self


class AggFunction(ABC):
    """Main class for all PostgreSQL aggregate function."""

    function_name: str

    def __init__(
        self: Self,
        *func_argument: SQLSelectable | Any,
        alias: str | None = None,
        **_kwargs: Any,
    ) -> None:
        """Initialize AggFunction.

        ### Parameters:
        - `func_argument`: arguments for the aggregate function.
        - `alias`: alias for the function result.
        - `kwargs`: any additional parameters for subclasses.
        """
        self.func_arguments: list[SQLSelectable | Any] = [
            *func_argument,
        ]
        self.alias = alias

    def querystring(self: Self) -> QueryString:
        """Build new `QueryString`.

        If function argument is `SQLSelectable`
        (usually it's a Field), then use `querystring()`
        method.

        If function argument isn't `SQLSelectable`,
        then try to convert this value into SQL-readable string.

        ### Returns:
        `QueryString` for aggregate function.
        """
        return QueryString(
            *self._querystring_args,
            sql_template=self.function_name + "(" + self._template_args + ")",
        )

    @property
    def _template_args(self: Self) -> str:
        return ", ".join(
            ["{}" for _ in self.func_arguments],
        )

    @property
    def _querystring_args(self: Self) -> list[str]:
        querystring_args: list[str] = []

        for func_argument in self.func_arguments:
            if isinstance(func_argument, SQLSelectable):
                querystring_args.append(
                    str(func_argument.querystring()),
                )
            else:
                querystring_args.append(
                    transform_value_to_sql(func_argument),
                )

        return querystring_args



from qaspen.sql_type.base import SQLType


class Json(SQLType):
    """Represent `JSON` PostgreSQL type in the python."""

    python_type = (dict, list)


class Jsonb(SQLType):
    """Represent `JSON` PostgreSQL type in the python."""

    python_type = (dict, list)


class Array(SQLType):
    """Represent `ARRAY` PostgreSQL type in the python."""

    python_type = list



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any, Final, Union, overload

from qaspen.base.operators import AllOperator, AnyOperator
from qaspen.exceptions import (
    FieldComparisonError,
    FieldDeclarationError,
    FieldValueValidationError,
)
from qaspen.fields import operators
from qaspen.fields.base import Field
from qaspen.fields.utils import MAX_STRING_FIELD_LENGTH, validate_max_length
from qaspen.qaspen_types import FieldDefaultType, FieldType
from qaspen.sql_type import primitive_types
from qaspen.statements.combinable_statements.filter_statement import Filter

if TYPE_CHECKING:
    from typing_extensions import Self


class BaseIntegerField(Field[int]):
    """Base field for all integer fields."""

    _available_comparison_types: tuple[type, ...] = (
        int,
        Field,
        AnyOperator,
        AllOperator,
    )
    _set_available_types: tuple[type, ...] = (int,)
    _available_max_value: int | None
    _available_min_value: int | None

    def __init__(
        self: Self,
        *pos_arguments: Any,
        is_null: bool = False,
        default: int | None = None,
        db_field_name: str | None = None,
        maximum: int | None = None,
        minimum: int | None = None,
    ) -> None:
        self._maximum: int | None = maximum
        self._minimum: int | None = minimum

        super().__init__(
            *pos_arguments,
            is_null=is_null,
            default=default,
            db_field_name=db_field_name,
        )

    def _validate_field_value(
        self: Self,
        field_value: int | None,
    ) -> None:
        """Validate field value.

        Check maximum and minimum values, if validation failed
        raise an error.

        :param field_value: new value for the field.

        :raises FieldValueValidationError: if value is too big or small.
        """
        super()._validate_field_value(
            field_value=field_value,
        )

        is_max_value_reached: Final = bool(
            field_value
            and self._available_max_value
            and field_value > self._available_max_value,
        )
        if is_max_value_reached:
            max_reached_err_msg: Final = (
                f"Field value - `{field_value}` "
                f"is greater than field `{self.__class__.__name__}` "
                f"can accommodate - `{self._available_max_value}`",
            )
            raise FieldValueValidationError(max_reached_err_msg)

        is_min_value_reached: Final = bool(
            field_value
            and self._available_min_value
            and field_value < self._available_min_value,
        )
        if is_min_value_reached:
            min_reached_err_msg: Final = (
                f"Field value - `{field_value}` "
                f"is less than field `{self.__class__.__name__}` "
                f"can accommodate - `{self._available_min_value}`",
            )
            raise FieldValueValidationError(min_reached_err_msg)
        if field_value and self._maximum and field_value > self._maximum:
            value_max_reached_err_msg: Final = (
                f"Field value - `{field_value}` is greater "
                f"than maximum you set `{self._maximum}`",
            )
            raise FieldValueValidationError(value_max_reached_err_msg)
        if field_value and self._minimum and field_value < self._minimum:
            value_min_reached_err_msg: Final = (
                f"Field value - `{field_value}` is less "
                f"than minimum you set `{self._minimum}`",
            )
            raise FieldValueValidationError(value_min_reached_err_msg)


class SmallIntField(BaseIntegerField):
    """SMALLINT field."""

    _available_max_value: int = 32767
    _available_min_value: int = -32768
    _sql_type = primitive_types.SmallInt


class IntegerField(BaseIntegerField):
    """INTEGER field."""

    _available_max_value: int = 2147483647
    _available_min_value: int = -2147483648
    _sql_type = primitive_types.Integer


class BigIntField(BaseIntegerField):
    """BIGINT field."""

    _available_max_value: int = 9223372036854775807
    _available_min_value: int = -9223372036854775808
    _sql_type = primitive_types.BigInt


class NumericField(BaseIntegerField):
    """NUMERIC field."""

    _sql_type = primitive_types.Numeric

    def __init__(
        self: Self,
        *pos_arguments: Any,
        precision: int | None = None,
        scale: int | None = None,
        is_null: bool = False,
        default: int | None = None,
        db_field_name: str | None = None,
        maximum: int | None = None,
        minimum: int | None = None,
    ) -> None:
        if not precision and scale:
            declaration_err_msg: Final = (
                "You cannot specify `scale` without `precision`.",
            )
            raise FieldDeclarationError(declaration_err_msg)

        super().__init__(
            *pos_arguments,
            is_null=is_null,
            default=default,
            db_field_name=db_field_name,
            maximum=maximum,
            minimum=minimum,
        )

        self.precision: int | None = precision
        self.scale: int | None = scale

    @property
    def _field_type(self: Self) -> str:
        field_type: str = self._sql_type.sql_type()
        if self.precision and self.scale:
            field_type += f"({self.precision}, {self.scale})"
        elif self.precision:
            field_type += f"({self.precision})"

        return field_type


class DecimalField(NumericField):
    """DECIMAL field.

    The same as `Numeric` field.
    """

    _sql_type = primitive_types.Decimal  # type: ignore[assignment]


class RealField(Field[Union[int, str]]):
    """REAL field."""

    _available_comparison_types: tuple[type, ...] = (
        str,
        float,
        Field,
        AnyOperator,
        AllOperator,
    )
    _set_available_types: tuple[type, ...] = (str, float)
    _sql_type = primitive_types.Real

    def __init__(
        self: Self,
        *pos_arguments: Any,
        is_null: bool = False,
        default: int | str | None = None,
        db_field_name: str | None = None,
    ) -> None:
        super().__init__(
            *pos_arguments,
            is_null=is_null,
            default=default,
            db_field_name=db_field_name,
        )


class DoublePrecisionField(Field[Union[int, str]]):
    """DOUBLE PRECISION field."""

    _available_comparison_types: tuple[type, ...] = (
        str,
        float,
        Field,
        AnyOperator,
        AllOperator,
    )
    _set_available_types: tuple[type, ...] = (str, float)
    _sql_type = primitive_types.DoublePrecision

    def __init__(
        self: Self,
        *pos_arguments: Any,
        is_null: bool = False,
        default: int | str | None = None,
        db_field_name: str | None = None,
    ) -> None:
        super().__init__(
            *pos_arguments,
            is_null=is_null,
            default=default,
            db_field_name=db_field_name,
        )


class BooleanField(Field[bool]):
    """BOOLEAN field."""

    _available_comparison_types: tuple[type, ...] = (
        bool,
        Field,
        AnyOperator,
        AllOperator,
    )
    _set_available_types: tuple[type, ...] = (bool,)
    _sql_type = primitive_types.Boolean

    def _validate_field_value(
        self: Self,
        field_value: bool | None,
    ) -> None:
        """Validate field value.

        Check all possible BOOLEAN field values.

        ### Parameters
        - field_value: field to validate

        ### Returns
        - `None`
        """
        super()._validate_field_value(
            field_value=field_value,
        )

        if field_value not in (True, False):
            validation_err_msg: Final = (
                f"Field value - `{field_value}` must be one of the following: "
                "True or False",
            )
            raise FieldValueValidationError(validation_err_msg)


class SerialBaseField(BaseIntegerField):
    """Base Serial field for all possible SERIAL fields."""

    def __init__(
        self: Self,
        *pos_arguments: Any,
        db_field_name: str | None = None,
        maximum: int | None = None,
        minimum: int | None = None,
        next_val_seq_name: str | None = None,
    ) -> None:
        """Create field with Serial signature.

        You can create name for the `next_val` sequence.
        Pass a name in the `next_val_seq_name` parameter
        in field creation.

        :param db_field_name: name for the field in the database.
        :param maximum: max number for the field at python level.
        :param minimum: min number for the field at python level.
        :param next_val_seq_name: name for the `nextval` sequence.
        """
        super().__init__(
            *pos_arguments,
            is_null=False,
            default=None,
            db_field_name=db_field_name,
            maximum=maximum,
            minimum=minimum,
        )

        self.next_val_seq_name: str | None = next_val_seq_name

    def _make_field_create_statement(self: Self) -> str:
        if self.next_val_seq_name:
            return (
                f"{self._original_field_name} "
                f"{self._sql_type.querystring()} NOT NULL "
                f"DEFAULT nextval('{self.next_val_seq_name}')"
            )
        return f"{self._original_field_name} {self._sql_type.querystring()}"


class SmallSerialField(SerialBaseField, SmallIntField):
    """SMALLSERIAL field.

    Its `SmallInt` field with `NOT NULL` property,
    and autoincrement functionality.
    """


class SerialField(SerialBaseField, IntegerField):
    """SERIAL field.

    Its `Integer` field with `NOT NULL` property,
    and autoincrement functionality.
    """

    _sub_field: str = "INTEGER"


class BigSerialField(SerialBaseField, BigIntField):
    """BIGSERIAL field.

    Its `BigInt` field with `NOT NULL` property,
    and autoincrement functionality.
    """

    _sub_field: str = "BIGINT"


AvailableComparisonTypes = (
    str,
    Field,
    AnyOperator,
    AllOperator,
)


class BaseStringField(Field[str]):
    """Base Field for all string-related Fields."""

    _available_comparison_types: tuple[type, ...] = AvailableComparisonTypes
    _set_available_types: tuple[type, ...] = (str,)

    @overload
    def __init__(
        self: Self,
        max_length: int = 255,
        is_null: bool = False,
        default: str | None = None,
        db_field_name: str | None = None,
    ) -> None:
        ...

    @overload
    def __init__(
        self: Self,
        is_null: bool = False,
        default: str | None = None,
        db_field_name: str | None = None,
    ) -> None:
        ...

    def __init__(
        self: Self,
        *args: Any,
        max_length: int = 255,
        is_null: bool = False,
        default: str | None = None,
        db_field_name: str | None = None,
    ) -> None:
        if max_length and validate_max_length(max_length=max_length):
            declaration_err_msg: Final = (
                f"`max_length` parameter must be less "
                f"or equal to {MAX_STRING_FIELD_LENGTH}",
            )
            raise FieldDeclarationError(declaration_err_msg)

        self._max_length: int = max_length

        super().__init__(
            *args,
            is_null=is_null,
            default=default,
            db_field_name=db_field_name,
        )

    def like(
        self: Self,
        comparison_value: str,
    ) -> Filter:
        """`LIKE` PostgreSQL clause.

        It allows you to use `LIKE` clause.

        ### Parameters:
        - `comparison_value`: value to compare with.

        ### Returns:
        Initialized `Filter`.
        """
        if isinstance(comparison_value, self._available_comparison_types):
            return Filter(
                field=self,
                comparison_value=comparison_value,
                operator=operators.LikeOperator,
            )

        comparison_err_msg: Final = (
            f"It's impossible to use `LIKE` operator "
            f"to compare {self.__class__.__name__} "
            f"and {type(comparison_value)}",
        )
        raise FieldComparisonError(comparison_err_msg)

    def not_like(
        self: Self,
        comparison_value: str,
    ) -> Filter:
        """`NOT LIKE` PostgreSQL clause.

        It allows you to use `NOT LIKE` clause.

        ### Parameters:
        - `comparison_value`: value to compare with.

        ### Returns:
        Initialized `Filter`.
        """
        if isinstance(comparison_value, self._available_comparison_types):
            return Filter(
                field=self,
                comparison_value=comparison_value,
                operator=operators.NotLikeOperator,
            )

        comparison_err_msg: Final = (
            f"It's impossible to use `NOT LIKE` operator "
            f"to compare {self.__class__.__name__} "
            f"and {type(comparison_value)}",
        )
        raise FieldComparisonError(comparison_err_msg)

    def ilike(
        self: Self,
        comparison_value: str,
    ) -> Filter:
        """`ILIKE` PostgreSQL clause.

        It allows you to use `ILIKE` clause.

        ### Parameters:
        - `comparison_value`: value to compare with.

        ### Returns:
        Initialized `Filter`.
        """
        if isinstance(comparison_value, self._available_comparison_types):
            return Filter(
                field=self,
                comparison_value=comparison_value,
                operator=operators.ILikeOperator,
            )
        comparison_err_msg: Final = (
            f"It's impossible to use `ILIKE` operator "
            f"to compare {self.__class__.__name__} "
            f"and {type(comparison_value)}",
        )
        raise FieldComparisonError(comparison_err_msg)

    def not_ilike(
        self: Self,
        comparison_value: str,
    ) -> Filter:
        """`NOT ILIKE` PostgreSQL clause.

        It allows you to use `NOT ILIKE` clause.

        ### Parameters:
        - `comparison_value`: value to compare with.

        ### Returns:
        Initialized `Filter`.
        """
        if isinstance(comparison_value, self._available_comparison_types):
            return Filter(
                field=self,
                comparison_value=comparison_value,
                operator=operators.NotILikeOperator,
            )

        comparison_err_msg: Final = (
            f"It's impossible to use `NOT ILIKE` operator "
            f"to compare {self.__class__.__name__} "
            f"and {type(comparison_value)}",
        )
        raise FieldComparisonError(comparison_err_msg)

    def _validate_field_value(
        self: Self,
        field_value: str | None,
    ) -> None:
        """Validate field value.

        If new value has length more that `max_length`, throw an error.

        :param field_value: new value for the field.

        :raises FieldValueValidationError: if the `max_length` is exceeded.
        """
        super()._validate_field_value(
            field_value=field_value,
        )

        if field_value and len(field_value) <= self._max_length:
            return
        if field_value and len(field_value) > self._max_length:
            validation_err_msg: Final = (
                f"You cannot set value with length {len(field_value)} "
                f"to the {self.__class__.__name__} with "
                f"`max_length` - {self._max_length}",
            )
            raise FieldValueValidationError(validation_err_msg)

    @property
    def _field_type(self: Self) -> str:
        return f"{self._sql_type.sql_type()}({self._max_length})"


class VarCharField(BaseStringField):
    """Varchar Field.

    Behave like normal PostgreSQL VARCHAR field.
    """

    _sql_type = primitive_types.VarChar


class TextField(Field[str]):
    """Text field.

    Behave like normal PostgreSQL TEXT field.
    """

    _available_comparison_types: tuple[type, ...] = AvailableComparisonTypes
    _set_available_types: tuple[type, ...] = (str,)
    _sql_type = primitive_types.Text


class CharField(Field[str]):
    """Char field.

    You cannot specify `max_length` parameter for this Field,
    it's always 1.

    If you want more characters, use `VarChar` field.
    """

    _available_comparison_types: tuple[type, ...] = AvailableComparisonTypes
    _set_available_types: tuple[type, ...] = (str,)
    _sql_type = primitive_types.Char

    def __init__(
        self: Self,
        *pos_arguments: Any,
        is_null: bool = False,
        default: str | None = None,
        db_field_name: str | None = None,
    ) -> None:
        super().__init__(
            *pos_arguments,
            is_null=is_null,
            default=default,
            db_field_name=db_field_name,
        )

    def _validate_field_value(
        self: Self,
        field_value: str | None,
    ) -> None:
        """Validate field value.

        If value length not equal 1 raise an error.

        :param field_value: new value for the field.

        :raises FieldValueValidationError: if value length not equal 1.
        """
        if not field_value or len(field_value) == 1:
            return

        validation_err_msg: Final = (
            f"CHAR field must always contain "
            f"only one character. "
            f"You tried to set {field_value}",
        )
        raise FieldValueValidationError(validation_err_msg)


class BaseDatetimeField(Field[FieldType]):
    """Base Field for all Date/Time fields."""

    _database_default: str = "CURRENT_DATE"

    def __init__(
        self: Self,
        *args: Any,
        is_null: bool = False,
        db_field_name: str | None = None,
        default: FieldDefaultType[FieldType] = None,
        database_default: bool = False,
    ) -> None:
        if default and database_default:
            declaration_err_msg: Final = (
                "Please specify either `default` or `database_default` "
                "for DateField.",
            )
            raise FieldDeclarationError(declaration_err_msg)

        super().__init__(
            *args,
            is_null=is_null,
            default=default,
            db_field_name=db_field_name,
        )

        self.database_default: Final = database_default

    @property
    def _field_default(self: Self) -> str:
        """Build DEFAULT string for a field.

        :returns: DEFAULT sql string for database.
        """
        if not self.database_default:
            return super()._field_default
        return f"DEFAULT {self._database_default}"


class BaseDateTimeFieldWithTZ(BaseDatetimeField[FieldType]):
    """Base Field for all Date/Time fields with TimeZone."""

    def __init__(
        self: Self,
        *args: Any,
        is_null: bool = False,
        db_field_name: str | None = None,
        default: FieldDefaultType[FieldType] = None,
        database_default: bool = False,
        with_timezone: bool = False,
    ) -> None:
        if default and database_default:
            declaration_err_msg: Final = (
                "Please specify either `default` or `database_default` "
                "for DateField.",
            )
            raise FieldDeclarationError(declaration_err_msg)

        super().__init__(
            *args,
            is_null=is_null,
            default=default,
            db_field_name=db_field_name,
            database_default=database_default,
        )

        self.with_timezone: Final = with_timezone

    @property
    def _field_type(self: Self) -> str:
        if self.with_timezone:
            return f"{self._sql_type.sql_type()} WITH TIME ZONE"
        return super()._field_type


class DateField(BaseDatetimeField[datetime.date]):
    """PostgreSQL type for `datetime.date` python type."""

    _available_comparison_types: tuple[
        type,
        ...,
    ] = (
        datetime.date,
        Field,
        AllOperator,
        AnyOperator,
    )
    _set_available_types: tuple[type, ...] = (datetime.date,)
    _sql_type = primitive_types.Date


class TimeField(BaseDateTimeFieldWithTZ[datetime.time]):
    """PostgreSQL type for `datetime.time` python type."""

    _database_default: str = "CURRENT_TIME"

    _available_comparison_types: tuple[
        type,
        ...,
    ] = (
        datetime.time,
        Field,
        AllOperator,
        AnyOperator,
    )
    _set_available_types: tuple[type, ...] = (datetime.time,)
    _sql_type = primitive_types.Time


class TimestampField(BaseDateTimeFieldWithTZ[datetime.datetime]):
    """PostgreSQL type for `datetime.datetime` python type."""

    _available_comparison_types: tuple[
        type,
        ...,
    ] = (
        datetime.datetime,
        Field,
        AllOperator,
        AnyOperator,
    )
    _set_available_types: tuple[type, ...] = (datetime.datetime,)
    _sql_type = primitive_types.Timestamp


class IntervalField(Field[datetime.timedelta]):
    """PostgreSQL type for `datetime.timedelta` python type."""

    _available_comparison_types: tuple[
        type,
        ...,
    ] = (
        datetime.timedelta,
        Field,
        AllOperator,
        AnyOperator,
    )
    _set_available_types: tuple[type, ...] = (datetime.timedelta,)
    _sql_type = primitive_types.Interval



"""Statement result for SelectStatement."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, List

from pydantic import TypeAdapter

from qaspen.statements.statement_result.result_variants import (
    MSGSpecStatementResult,
    PydanticStatementResult,
    RawStatementResult,
)

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.qaspen_types import MSGSpecStruct, PydanticModel


class SelectStatementResult(
    RawStatementResult,
    PydanticStatementResult,
    MSGSpecStatementResult,
):
    """Result for select statement."""

    def __init__(
        self: Self,
        engine_result: list[dict[str, Any]],
    ) -> None:
        """Initialize `SelectStatementResult`.

        It is used to get results after statement execution.

        ### Parameters:
        - `engine_result`: result from the engine.
        - `pydantic_model`: pydantic model for engine result.
        """
        super().__init__(engine_result=engine_result)

    def result(
        self: Self,
    ) -> list[dict[str, Any]]:
        """Return result as-is from engine."""
        return self._engine_result

    def to_pydantic(
        self: Self,
        pydantic_model: type[PydanticModel],
    ) -> list[PydanticModel]:
        """Return result as a pydantic model.

        You can pass the pydantic model in the method and
        result will be transformed into you pydantic model.

        This passed pydantic model will override pydantic model
        that you passed when building statement.

        ### Parameters:
        - `pydantic_model`: pydantic model for engine result.

        ### Returns:
        list of `pydantic_models`.
        """
        type_adapter = TypeAdapter(
            List[pydantic_model],  # type: ignore[valid-type]
        )
        return type_adapter.validate_python(self._engine_result)

    def to_msgspec(
        self: Self,
        msgspec_struct: type[MSGSpecStruct],
    ) -> list[MSGSpecStruct]:
        """Return result as a list of msgspec structs.

        You can pass the msgspec struct type in the method and
        result will be transformed into you pydantic model.

        ### Parameters:
        - `msgspec_struct`: msgspec struct for engine result.

        ### Returns:
        list of `msgspec_struct`.
        """
        return [
            msgspec_struct(**single_result)
            for single_result in self._engine_result
        ]



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Final,
    Generator,
    Generic,
    Iterable,
    overload,
)

from qaspen.aggregate_functions.base import AggFunction
from qaspen.fields.aliases import FieldAliases
from qaspen.fields.base import Field
from qaspen.qaspen_types import FromTable
from qaspen.querystring.querystring import QueryString
from qaspen.statements.base import Executable
from qaspen.statements.combinable_statements.filter_statement import (
    FilterStatement,
)
from qaspen.statements.combinable_statements.join_statement import (
    Join,
    JoinStatement,
    JoinType,
)
from qaspen.statements.combinable_statements.order_by_statement import (
    OrderByStatement,
)
from qaspen.statements.statement import BaseStatement
from qaspen.statements.statement_result.select_statement_result import (
    SelectStatementResult,
)
from qaspen.statements.sub_statements.limit_statement import LimitStatement
from qaspen.statements.sub_statements.offset_statement import OffsetStatement

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.abc.db_engine import BaseEngine
    from qaspen.clauses.order_by import OrderBy
    from qaspen.statements.combinable_statements.combinations import (
        CombinableExpression,
    )
    from qaspen.statements.exists_statement import ExistsStatement
    from qaspen.statements.intersect_statement import IntersectStatement
    from qaspen.statements.union_statement import UnionStatement
    from qaspen.table.base_table import BaseTable


class SelectStatement(
    BaseStatement,
    Executable[SelectStatementResult],
    Generic[FromTable],
):
    """Main entry point for all SELECT queries.

    You shouldn't create instance of this class by yourself,
    instead you must use Table class.

    Example:
    -------
    ```
    class Buns(BaseTable, table_name="buns"):
        name: VarCharField = VarCharField()


    select_statement: SelectStatement = Buns.select()
    ```
    """

    def __init__(
        self: Self,
        from_table: type[FromTable],
        select_objects: Iterable[Field[Any] | AggFunction],
    ) -> None:
        self._from_table: Final[type[FromTable]] = from_table

        self._select_fields: Iterable[Field[Any]] = []
        self._select_agg_functions: Iterable[AggFunction] = []

        for select_object in select_objects:
            if isinstance(select_object, Field):
                self._select_fields.append(select_object)
            elif isinstance(select_object, AggFunction):
                self._select_agg_functions.append(
                    select_object,
                )

        self.final_select_fields: list[Field[Any]] = []
        self.exist_prefixes: Final[list[str]] = []

        self._filter_statement: FilterStatement = FilterStatement()
        self._limit_statement: LimitStatement = LimitStatement()
        self._offset_statement: OffsetStatement = OffsetStatement()
        self._order_by_statement: OrderByStatement = OrderByStatement()
        self._join_statement: JoinStatement = JoinStatement()
        self._field_aliases: FieldAliases = FieldAliases()

    def __await__(
        self: Self,
    ) -> Generator[None, None, SelectStatementResult]:
        """SelectStatement can be awaited.

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()

        async def main() -> None:
            list_of_buns = await Buns.select()
            print(list_of_buns)
        ```
        """
        engine: BaseEngine[
            Any,
            Any,
            Any,
        ] | None = self._from_table._table_meta.database_engine
        if not engine:
            engine_err_msg: Final = "There is no database engine."
            raise AttributeError(engine_err_msg)

        return self.execute(engine=engine).__await__()

    async def execute(
        self: Self,
        engine: BaseEngine[Any, Any, Any],
    ) -> SelectStatementResult:
        """Execute select statement.

        This is manual execution.
        You can pass specific engine and set if you want
        to retrieve table objects instead of list of dicts.

        ### Parameters
        :param engine: subclass of BaseEngine.

        ### Returns
        :returns: list of dicts or list of table instances.
        """
        raw_query_result: list[dict[str, Any]] = await engine.execute(
            querystring=self.querystring(),
            fetch_results=True,
        )

        return SelectStatementResult(
            engine_result=raw_query_result,
        )

    def where(
        self: Self,
        *where_arguments: CombinableExpression,
    ) -> Self:
        """Add where clause to the statement.

        It's possible to use this method as many times as you want.
        If you use `where` more than one time, clauses will be connected
        with `AND` operator.

        Fields have different methods for the comparison.
        Also, you can pass the combination of the `where` clauses.

        Below you will see easy and advanced examples.

        One Where Example:
        ------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()

        statement = Buns.select().where(
            Buns.name == "Tasty",
        )
        ```
        Combination Where Example:
        ------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()
            example_field: VarCharField = VarCharField()

        statement = Buns.select().where(
            (Buns.name == "Tasty")
            & (Buns.description != "Not Tasty")
            | Buns.example_field.eq("Real Example"))
        )
        ```
        In this statement `&` is AND, `|` is OR.
        """
        self._filter_statement.add_filter(*where_arguments)
        return self

    def limit(
        self: Self,
        limit: int,
    ) -> Self:
        """Set limit to the query.

        If you use this method more than 1 time, previous result
        will be overridden.

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()


        statement = Buns.select().limit(limit=10)

        # if you will use limit() again, previous number
        # will be overridden.

        statement = statement.limit(limit=30)

        # not limit is LIMIT 30.
        ```
        """
        self._limit_statement.limit(limit_number=limit)
        return self

    def offset(
        self: Self,
        offset: int,
    ) -> Self:
        """Set offset to the query.

        If you use this method more than 1 time, previous result
        will be overridden.

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()


        statement = Buns.select().offset(offset=10)

        # if you will use offset() again, previous number
        # will be overridden.

        statement = statement.offset(offset=30)

        # not limit is OFFSET 30.
        ```
        """
        self._offset_statement.offset(offset_number=offset)
        return self

    def limit_offset(
        self: Self,
        limit: int,
        offset: int,
    ) -> Self:
        """Set offset and limit at the same time.

        See `limit()` and `offset()` methods description.
        """
        self._limit_statement.limit(limit_number=limit)
        self._offset_statement.offset(offset_number=offset)
        return self

    def order_by(
        self: Self,
        field: Field[Any] | None = None,
        ascending: bool = True,
        nulls_first: bool = True,
        order_bys: Iterable[OrderBy] | None = None,
    ) -> Self:
        """Set ORDER BY.

        You can specify `field`, `ascending` and `nulls_first`
        parameters and/or pass `order_bys` with OrderBy instances.

        ### Parameters
        :param field: subclass of BaseField.
        :param ascending: if True then `ASC` else `DESC`.
        :param nulls_first: if True then `NULLS FIRST` else `NULLS LAST`.
        :param order_bys: list of instances of order by.

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()


        # first variant
        statement = Buns.select().order_by(
            field=Buns.name,
            # it's not necessary, but you can specify.
            ascending=False,  # by default its `True`
            nulls_first=False,  # by default its `True`
        )

        # second variant
        order_bys = [
            OrderBy(
                field=Buns.name,
                # it's not necessary, but you can specify.
                ascending=False,  # by default its `True`
                nulls_first=False,  # by default its `True`
            ),
            OrderBy(
                field=Buns.surname,
            ),
        ]
        statement = Buns.select().order_by(
            order_bys=order_bys,
        )
        ```
        """
        if field:
            self._order_by_statement.order_by(
                field=field,
                ascending=ascending,
                nulls_first=nulls_first,
            )
        if order_bys:
            self._order_by_statement.order_by(
                order_by_statements=order_bys,
            )
        return self

    @overload
    def union(
        self: Self,
        union_with: SelectStatement[FromTable],
        union_all: bool = False,
    ) -> UnionStatement[FromTable]:
        ...

    @overload
    def union(
        self: Self,
        union_with: SelectStatement[Any],
        union_all: bool = False,
    ) -> UnionStatement[FromTable]:
        ...

    def union(
        self: Self,
        union_with: SelectStatement[FromTable],
        union_all: bool = False,
    ) -> UnionStatement[FromTable]:
        """Create union statement.

        Combines two `SelectStatement`'s and creates
        `UnionStatement`.

        ### Parameters
        :param union_with: `SelectStatement` for union.
        :param union_all: if True than `UNION ALL` else `UNION`.

        ### Returns
        :returns: `UnionStatement`.

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()


        statement_1 = Buns.select()
        statement_2 = Buns.select()
        union = statement_1.union(
            union_with=statement_2,
            union_all=False,
        )
        ```
        """
        from qaspen.statements.union_statement import UnionStatement

        return UnionStatement[FromTable](
            left_expression=self,
            right_expression=union_with,
            union_all=union_all,
        )

    def intersect(
        self: Self,
        intersect_with: SelectStatement[FromTable],
    ) -> IntersectStatement:
        """Create intersect statement.

        Combines two `SelectStatement`'s and creates
        `IntersectStatement`

        ### Parameters
        :param intersect_with: `SelectStatement` for intersect.

        ### Returns
        :returns: `IntersectStatement`.

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()


        statement_1 = Buns.select()
        statement_2 = Buns.select()
        intersect = statement_1.intersect(
            intersect_with=statement_2,
        )
        ```
        """
        from qaspen.statements.intersect_statement import IntersectStatement

        return IntersectStatement(
            left_expression=self,
            right_expression=intersect_with,
        )

    def exists(self: Self) -> ExistsStatement:
        """Create exists statement.

        Create exists statement that return True or False.

        ### Returns
        :returns: `ExistsStatement`.

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()


        exists_statement = Buns.select().where(
            Buns.name == "CallMe",
        ).exists()
        ```
        """
        from qaspen.statements.exists_statement import ExistsStatement

        self._select_fields = []
        for join in self._join_statement.join_expressions:
            join._fields = []
        return ExistsStatement(
            select_statement=self,
        )

    def join(
        self: Self,
        join_table: type[BaseTable],
        based_on: CombinableExpression,
    ) -> Self:
        """Add `JOIN` to the SelectStatement.

        Set `ON` condition, it can be a combination.

        ### Parameters
            - `join_table`: table in JOIN.
            - `based_on`: Field's comparisons.

        ### Returns
        `self`

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()


        class Cookies(BaseTable, table_name="cookies"):
            bun_name: VarCharField = VarCharField()
            filling: VarCharField = VarCharField()
            topping: VarCharField = VarCharField()

        statement = (
            Buns
            .select(
                Cookies.filling,
                Cookies.topping,
            )
            .join(
                join_table=Cookies,
                based_on=Buns.name == Cookies.bun_name,
            )
        )
        ```

        :returns: `SelectStatement`
        """
        return self._join_on(
            join_table=join_table,
            based_on=based_on,
            join_type=JoinType.JOIN,
        )

    def inner_join(
        self: Self,
        join_table: type[BaseTable],
        based_on: CombinableExpression,
    ) -> Self:
        """Add `INNER JOIN` to the SelectStatement.

        Set `ON` condition, it can be a combination.

        ### Parameters
        - `join_table`: table in JOIN.
        - `based_on`: Field's comparisons.

        ### Returns
        `self`

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()


        class Cookies(BaseTable, table_name="cookies"):
            bun_name: VarCharField = VarCharField()
            filling: VarCharField = VarCharField()
            topping: VarCharField = VarCharField()

        statement = (
            Buns
            .select(
                Cookies.filling,
                Cookies.topping,
            )
            .inner_join(
                join_table=Cookies,
                based_on=Buns.name == Cookies.bun_name,
            )
        )
        ```

        :returns: `SelectStatement`
        """
        return self._join_on(
            join_table=join_table,
            based_on=based_on,
            join_type=JoinType.INNERJOIN,
        )

    def left_join(
        self: Self,
        join_table: type[BaseTable],
        based_on: CombinableExpression,
    ) -> Self:
        """Add `LEFT JOIN` to the SelectStatement.

        Set `ON` condition, it can be a combination.

        ### Parameters
        - `join_table`: table in JOIN.
        - `based_on`: Field's comparisons.

        ### Returns
        `self`

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()


        class Cookies(BaseTable, table_name="cookies"):
            bun_name: VarCharField = VarCharField()
            filling: VarCharField = VarCharField()
            topping: VarCharField = VarCharField()

        statement = (
            Buns
            .select(
                Cookies.filling,
                Cookies.topping
            )
            .left_join(
                join_table=Cookies,
                based_on=Buns.name == Cookies.bun_name,
            )

        )
        ```

        :returns: `SelectStatement`
        """
        return self._join_on(
            join_table=join_table,
            based_on=based_on,
            join_type=JoinType.LEFTJOIN,
        )

    def right_join(
        self: Self,
        join_table: type[BaseTable],
        based_on: CombinableExpression,
    ) -> Self:
        """Add `RIGHT JOIN` to the SelectStatement.

        Set `ON` condition, it can be a combination.

        ### Parameters
        - `join_table`: table in JOIN.
        - `based_on`: Field's comparisons.

        ### Returns
        `self`

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()


        class Cookies(BaseTable, table_name="cookies"):
            bun_name: VarCharField = VarCharField()
            filling: VarCharField = VarCharField()
            topping: VarCharField = VarCharField()

        statement = (
            Buns
            .select(
                Cookies.filling,
                Cookies.topping,
            )
            .right_join(
                join_table=Cookies,
                based_on=Buns.name == Cookies.bun_name,
            )
        )
        ```

        :returns: `SelectStatement`
        """
        return self._join_on(
            join_table=join_table,
            based_on=based_on,
            join_type=JoinType.RIGHTJOIN,
        )

    def full_outer_join(
        self: Self,
        join_table: type[BaseTable],
        based_on: CombinableExpression,
    ) -> Self:
        """Add `FULL OUTER JOIN` to the SelectStatement.

        You can specify what fields you want to get from
        the joined table.
        And set `ON` condition, it can be a combination.

        ### Parameters
        - `join_table`: table in JOIN.
        - `based_on`: Field's comparisons.

        ### Returns
        `self`

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()


        class Cookies(BaseTable, table_name="cookies"):
            bun_name: VarCharField = VarCharField()
            filling: VarCharField = VarCharField()
            topping: VarCharField = VarCharField()


        statement = (
            Buns
            .select(
                Cookies.filling,
                Cookies.topping,
            )
            .full_outer_join(
                join_table=Cookies,
                based_on=Buns.name == Cookies.bun_name,
            )
        )
        ```

        :returns: `SelectStatement`
        """
        return self._join_on(
            join_table=join_table,
            based_on=based_on,
            join_type=JoinType.FULLOUTERJOIN,
        )

    def add_join(
        self: Self,
        *join: Join,
    ) -> Self:
        """Add one ore more joins to the SelectStatement.

        You can create JOINs instances by yourself and pass here.

        Example:
        -------
        ```
        class Computer(BaseTable, table_name="computer"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()


        class Processor(BaseTable, table_name="processor"):
            computer_name: VarCharField = VarCharField()
            processor_name: VarCharField = VarCharField()


        class Architecture(BaseTable, table_name="Architecture")
            title: VarCharField = VarCharField()
            processor_name: VarCharField = VarCharField()


        processor_join = Join(
            fields=[Processor.processor_name],
            from_table=Computer,
            join_table=Processor,
            on=Computer.name == Processor.computer_name,
            # You must control join aliases.
            join_alias="processor_join",
        )

        architecture_join = Join(
            fields=[Architecture.title],
            from_table=Computer,
            join_table=Architecture,
            on=processor_join.processor_name == Architecture.processor_name,
            join_alias="architecture_join",
        )

        statement = Computer.select().add_join(
            processor_join,
            architecture_join,
        )
        ```
        """
        self._join_statement.add_join(
            *join,
        )
        return self

    def querystring(self: Self) -> QueryString:
        """Build querystring.

        Can be transformed into the SQL query with `str` method.

        :returns: QueryString.
        """
        sql_querystring = self._select_from()
        sql_querystring += self._join_statement.querystring()
        sql_querystring += self._filter_statement.querystring()
        sql_querystring += self._order_by_statement.querystring()
        sql_querystring += self._limit_statement.querystring()
        sql_querystring += self._offset_statement.querystring()

        return sql_querystring

    def _join_on(
        self: Self,
        join_table: type[BaseTable],
        based_on: CombinableExpression,
        join_type: JoinType,
    ) -> Self:
        self._join_statement.join(
            join_table=join_table,
            from_table=self._from_table,
            on=based_on,
            join_type=join_type,
        )
        return self

    def _prepare_fields(
        self: Self,
    ) -> list[Field[Any]]:
        if self.final_select_fields:
            return self.final_select_fields
        final_select_fields: Final[list[Field[Any]]] = []

        fields_to_select: list[Field[Any]] = []
        fields_to_select.extend(self._prepare_select_fields())
        fields_to_select.extend(
            self._join_statement._retrieve_all_join_fields(),
        )

        for field in fields_to_select:
            aliased_field = self._field_aliases.add_alias(
                field=field,
            )
            final_select_fields.append(aliased_field)

        self.final_select_fields = final_select_fields
        return final_select_fields

    def _select_from(self: Self) -> QueryString:
        to_select_fields: str = ", ".join(
            [field.field_name for field in self._prepare_fields()],
        )
        to_select_agg_funcs = ", ".join(
            [
                str(agg_func.querystring())
                for agg_func in self._select_agg_functions
            ],
        )

        if to_select_fields and to_select_agg_funcs:
            final_select_objects = f"{to_select_fields}, {to_select_agg_funcs}"
        elif to_select_fields and not to_select_agg_funcs:
            final_select_objects = to_select_fields
        elif to_select_agg_funcs and not to_select_fields:
            final_select_objects = to_select_agg_funcs

        if self._from_table._table_meta.alias:
            return QueryString(
                final_select_objects or "1",
                self._from_table.schemed_original_table_name(),
                self._from_table._table_meta.alias,
                sql_template="SELECT {} FROM {} as {}",
            )

        return QueryString(
            final_select_objects or "1",
            self._from_table.schemed_original_table_name(),
            sql_template="SELECT {} FROM {}",
        )

    def _prepare_select_fields(
        self: Self,
    ) -> list[Field[Any]]:
        fields_from_original_table: list[Field[Any]] = []
        fields_from_joined_table: dict[
            str,
            list[Field[Any]],
        ] = {}

        for field in self._select_fields:
            is_original = field._field_data.from_table == self._from_table
            if is_original:
                fields_from_original_table.append(
                    field,
                )
            else:
                fields_from_joined_table.setdefault(
                    field._field_data.from_table.original_table_name(),
                    [],
                ).append(field)

        for join_expr in self._join_statement.join_expressions:
            join_fields = fields_from_joined_table.get(
                join_expr._join_table.original_table_name(),
            )
            if not join_fields:
                continue
            join_expr.add_fields(join_fields)

        return fields_from_original_table



"""Base SQL operators."""
from typing import Final

from typing_extensions import Self

from qaspen.base.sql_base import SQLSelectable
from qaspen.querystring.querystring import QueryString


class AnyOperator:
    """ANY PostgreSQL operator.

    Provide functionality of ANY PostgreSQL operator.
    """

    def __init__(
        self: Self,
        subquery: SQLSelectable,
    ) -> None:
        """Initialize AnyOperator.

        ### Parameters:
        - `subquery`: Any object that provides `querystring()` method.

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()


        select_statement = (
            Buns
            .select()
            .where(
                Buns.name == AnyOperator(
                    subquery=Buns.select()
                )
            )
        )
        ```
        """
        self.subquery: Final = subquery

    def querystring(self: Self) -> QueryString:
        """Build `QueryString` object.

        ### Returns:
        `QueryString`
        """
        sql_template: str = (
            "ANY (" + self.subquery.querystring().sql_template + ")"
        )
        return QueryString(
            *self.subquery.querystring().template_arguments,
            sql_template=sql_template,
        )


class AllOperator:
    """ALL PostgreSQL operator.

    Provide functionality of ALL PostgreSQL operator.
    """

    def __init__(
        self: Self,
        subquery: SQLSelectable,
    ) -> None:
        """Initialize AllOperator.

        ### Parameters:
        - `subquery`: Any object that provides `querystring()` method.

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()


        select_statement = (
            Buns
            .select()
            .where(
                Buns.name == AllOperator(
                    subquery=Buns.select()
                )
            )
        )
        ```
        """
        self.subquery: Final = subquery

    def querystring(self: Self) -> QueryString:
        """Build `QueryString` object.

        ### Returns:
        `QueryString`
        """
        sql_template: str = (
            "ALL (" + self.subquery.querystring().sql_template + ")"
        )
        return QueryString(
            *self.subquery.querystring().template_arguments,
            sql_template=sql_template,
        )



from __future__ import annotations

import contextvars
import typing
from abc import ABC, abstractmethod

from qaspen.abc.abc_types import (
    DBConnection,
    EngineConnectionPool,
    EngineTransaction,
)
from qaspen.utils.engine_utils import parse_database

if typing.TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.querystring.querystring import QueryString


class BaseEngine(
    ABC,
    typing.Generic[
        DBConnection,
        EngineConnectionPool,
        EngineTransaction,
    ],
):
    """Base engine class for all possible engines."""

    engine_type: str

    def __init__(
        self: Self,
        connection_url: str,
        **_kwargs: typing.Any,
    ) -> None:
        """Initialize Engine.

        ### Parameters:
        - `kwargs`: just for inheritance, subclasses won't
            have problems with type hints.
        """
        self.running_transaction: contextvars.ContextVar[
            EngineTransaction | None,
        ] = contextvars.ContextVar(
            "running_transaction",
            default=None,
        )
        self.connection_url = connection_url

    @typing.overload
    async def execute(  # type: ignore[misc]
        self: Self,
        querystring: QueryString,
        in_pool: bool = True,
        fetch_results: typing.Literal[True] = True,
        **_kwargs: typing.Any,
    ) -> list[dict[str, typing.Any]]:
        ...

    @typing.overload
    async def execute(
        self: Self,
        querystring: QueryString,
        in_pool: bool = True,
        fetch_results: typing.Literal[False] = False,
        **_kwargs: typing.Any,
    ) -> None:
        ...

    @abstractmethod
    async def execute(
        self: Self,
        querystring: QueryString,
        in_pool: bool = True,
        fetch_results: bool = True,
        **_kwargs: typing.Any,
    ) -> list[dict[str, typing.Any]] | None:
        """Execute a querystring.

        Run querystring and return raw result as in
        database driver.

        ### Parameters:
        - `querystring`: `QueryString` or it's subclasses.
        - `in_pool`: execution in connection pool
            or in a new connection.
        - `kwargs`: just for inheritance, subclasses won't
            have problems with type hints.

        ### Returns:
        Raw result from database driver.
        """

    @abstractmethod
    async def prepare_database(
        self: Self,
    ) -> None:
        """Prepare database.

        typing.Anything that must be done before database will be
        ready to execute queries.

        For example, create extensions in PostgreSQL.
        """

    @abstractmethod
    async def create_connection_pool(
        self: Self,
    ) -> EngineConnectionPool:
        """Create new connection pool.

        If connection pool already exists return it.

        ### Returns:
        Connection pool from a database driver.
        """

    @abstractmethod
    async def stop_connection_pool(
        self: Self,
    ) -> None:
        """Close connection pool.

        If connection pool doesn't exist, print warning.
        """

    @abstractmethod
    async def connection(
        self: Self,
    ) -> DBConnection:
        """Create new connection.

        Build new connection, don't take one from connection pool.

        ### Returns:
        Connection from a database driver.
        """

    @abstractmethod
    def transaction(self: Self) -> EngineTransaction:
        """Create new transaction.

        Create new transaction.
        Transaction must support async context manager protocol.

        ### Returns:
        Transaction for this engine.
        """

    @property
    def database(self: Self) -> str:
        """Get database from connection url.

        ### Returns:
        Connection from connection url.
        """
        return parse_database(self.connection_url)



from __future__ import annotations

import abc
import dataclasses
from typing import TYPE_CHECKING

from qaspen.fields.operators import (
    ANDOperator,
    BaseOperator,
    NotOperator,
    OROperator,
)
from qaspen.querystring.querystring import QueryString

if TYPE_CHECKING:
    from typing_extensions import Self


class CombinableExpression(abc.ABC):
    """Base class for all classes that can be combined.

    It provides base realization for `add`, `or` and `not`.

    Usually used for `Filter` classes, so you can combine
    filter limitless.
    """

    @abc.abstractmethod
    def querystring(self: Self) -> QueryString:
        """Build new querystring for this expression."""
        ...

    def __and__(
        self: Self,
        expression: CombinableExpression,
    ) -> ANDExpression:
        """Add new expression to exist one.

        It creates `ANDExpression`, that means that now
        two expression combined with `AND` SQL clause.

        ### Parameters:
        - `expression`: other combinable expression.

        ### Returns:
        `ANDExpression`
        """
        return ANDExpression(
            left_expression=self,
            right_expression=expression,
        )

    def __or__(
        self: Self,
        expression: CombinableExpression,
    ) -> ORExpression:
        """Combine two expressions into new one.

        It creates `ORExpression`, that means that now
        two expression combined with `OR` SQL clause.

        ### Parameters:
        - `expression`: other combinable expression.

        ### Returns:
        `ORExpression`
        """
        return ORExpression(
            left_expression=self,
            right_expression=expression,
        )

    def __invert__(self: Self) -> NotExpression:
        """Create invert expression.

        It creates `NotExpression`, that means that now
        expression will reverse the condition.

        ### Parameters:
        - `expression`: other combinable expression.

        ### Returns:
        `NotExpression`
        """
        return NotExpression(
            left_expression=self,
            right_expression=None,  # type: ignore[arg-type]
        )


@dataclasses.dataclass
class ExpressionsCombination(CombinableExpression):
    """Combination of CombinableExpressions.

    This class contains unlimited number of
    `CombinableExpression`, can aggregate them,
    and create one QueryString.
    """

    left_expression: CombinableExpression
    right_expression: CombinableExpression
    operator: type[BaseOperator] = BaseOperator

    def querystring(self: Self) -> QueryString:
        """Build new single `QueryString`.

        We recursively call `querystring` method
        for all `CombinableExpression` and create single
        QueryString.

        ### Returns:
        New `QueryString`
        """
        return QueryString(
            self.left_expression.querystring(),
            self.right_expression.querystring(),
            sql_template="{} " + self.operator.operation_template + " {}",
        )


@dataclasses.dataclass
class ANDExpression(ExpressionsCombination):
    """Expression for `AND` PostgreSQL clause."""

    operator: type[ANDOperator] = ANDOperator


@dataclasses.dataclass
class ORExpression(ExpressionsCombination):
    """Expression for `OR` PostgreSQL clause."""

    operator: type[OROperator] = OROperator


@dataclasses.dataclass
class NotExpression(ExpressionsCombination):
    """Expression for `NOT` PostgreSQL clause."""

    operator: type[NotOperator] = NotOperator

    def querystring(self: Self) -> QueryString:
        """Build QueryString."""
        return QueryString(
            self.left_expression.querystring(),
            sql_template=self.operator.operation_template,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Final, cast

from qaspen.exceptions import QueryResultLookupError
from qaspen.querystring.querystring import QueryString
from qaspen.statements.base import Executable
from qaspen.statements.combinable_statements.combinations import (
    CombinableExpression,
)
from qaspen.statements.statement import BaseStatement

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.abc.db_engine import BaseEngine
    from qaspen.qaspen_types import FromTable
    from qaspen.statements.select_statement import SelectStatement


class ExistsStatement(
    BaseStatement,
    CombinableExpression,
    Executable[bool],
):
    """Statement for `Exists` Statement."""

    def __init__(
        self: Self,
        select_statement: SelectStatement[FromTable],
    ) -> None:
        self._select_statement: Final[
            SelectStatement[FromTable],
        ] = select_statement

    def __await__(self: Self) -> Any:
        """Make ExistsStatement awaitable."""
        engine: BaseEngine[
            Any,
            Any,
            Any,
        ] | None = (
            self._select_statement._from_table._table_meta.database_engine
        )
        if not engine:
            engine_err_msg: Final = "There is no database engine."
            raise AttributeError(engine_err_msg)

        return self.execute(engine=engine).__await__()

    def querystring(self: Self) -> QueryString:
        """Return querystring for comparisons.

        Used in methods like `Field.contains(querystring=...)`.
        """
        return QueryString(
            self._select_statement.querystring(),
            sql_template="EXISTS ({})",
        )

    def querystring_for_select(self: Self) -> QueryString:
        """Create querystring for SELECT."""
        return QueryString(
            self._select_statement.querystring(),
            sql_template="SELECT EXISTS ({})",
        )

    async def execute(
        self: Self,
        engine: BaseEngine[Any, Any, Any],
    ) -> bool:
        """Execute Exists statement.

        You can pass custom engine in this method.

        :param engine: subclass of BaseEngine.
        """
        raw_query_result: list[dict[str, Any]] = await engine.execute(
            querystring=self.querystring_for_select(),
            fetch_results=True,
        )

        try:
            return cast(
                bool,
                raw_query_result[0]["exists"],
            )
        except LookupError as exc:
            lookup_err_msg: Final = (
                "Cannot get result for ExistsStatement. "
                "Please check your statement.",
            )
            raise QueryResultLookupError(
                lookup_err_msg,
            ) from exc



from __future__ import annotations

import copy
import inspect
import typing

from qaspen.fields.base import Field
from qaspen.statements.select_statement import SelectStatement
from qaspen.table.meta_table import MetaTable

if typing.TYPE_CHECKING:
    from qaspen.aggregate_functions.base import AggFunction
    from qaspen.qaspen_types import FieldType

T_ = typing.TypeVar(
    "T_",
    bound="BaseTable",
)


class BaseTable(MetaTable, abstract=True):
    """Main Table in the `Qaspen` library.

    Any Table must be created with this class as base class.
    """

    @classmethod
    def select(
        cls: type[T_],
        *select: Field[typing.Any] | AggFunction,
    ) -> SelectStatement[T_]:
        """Create SelectStatement based on table.

        :param select_fields: fields to select.
            By default select all possible fields.

        :returns: SelectStatement.
        """
        select_statement: typing.Final[SelectStatement[T_]] = SelectStatement(
            select_objects=select or cls.all_fields(),
            from_table=cls,
        )
        return select_statement

    @classmethod
    def all_fields(cls: type[T_]) -> list[Field[FieldType]]:
        """Return all fields in the table.

        :returns: list of fields.
        """
        return [*cls._table_meta.table_fields.values()]

    @classmethod
    def aliased(cls: type[T_], alias: str) -> type[T_]:
        """Create aliased version of the Table.

        ### Parameters
        :param alias: alias to the field.

        ### Returns
        :returns: Same Table, but aliased.

        Example:
        -------
        ```python
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()
            description: VarCharField = VarCharField()


        # It has the same type with all autosuggestions.
        AliasedBuns = Buns.aliased(alias="PrimaryBuns")
        ```
        """
        return cls._aliased(alias=alias)

    @classmethod
    def _aliased(cls: type[T_], alias: str) -> type[T_]:
        """Add alias to the table.

        We must create new `aliased_table` because in other
        way `alias` change will affect all other queries where
        aliased table is used.

        It'll be used in queries instead of real table name.

        :param alias: string alias to the table.

        :returns: the same table but with the alias.
        """
        aliased_table = type(
            cls.__name__,
            (cls,),
            {},
        )

        attributes = inspect.getmembers(
            cls,
            lambda member: not (inspect.isroutine(member)),
        )
        only_field_attributes: dict[str, Field[typing.Any]] = {
            attribute[0]: copy.deepcopy(attribute[1])
            for attribute in attributes
            if issubclass(type(attribute[1]), Field)
        }

        for table_param_name, table_param in cls.__dict__.items():
            setattr(aliased_table, table_param_name, table_param)

        for field_name, field in only_field_attributes.items():
            field._field_data.from_table = aliased_table
            setattr(aliased_table, field_name, field)

        aliased_table._table_meta = (  # type: ignore[attr-defined]
            copy.deepcopy(
                cls._table_meta,
            )
        )
        aliased_table._table_meta.alias = alias  # type: ignore[attr-defined]

        table_meta_fields = (
            aliased_table._table_meta.table_fields.values()  # type: ignore[attr-defined]
        )
        for field in table_meta_fields:
            field._field_data.from_table = aliased_table

        return aliased_table

    @classmethod
    def is_aliased(cls: type[T_]) -> bool:
        """Return flag that says does the table have an alias.

        :returns: boolean flag.
        """
        return bool(cls._table_meta.alias)

    @classmethod
    def original_table_name(cls: type[T_]) -> str:
        """Return original name of the table.

        Without aliases and other modifications.

        ### Returns:
        Table name as a string.
        """
        return cls._table_meta.table_name

    @classmethod
    def table_name(cls: type[T_]) -> str:
        """Return the original table name or alias to it.

        :returns: original table name or alias.
        """
        return cls._table_meta.alias or cls.original_table_name()

    @classmethod
    def schemed_table_name(cls: type[T_]) -> str:
        """Return the original table name or alias to it with schema.

        :returns: original table name or alias with corresponding schema.
        """
        return f"{cls._table_meta.table_schema}.{cls.table_name()}"

    @classmethod
    def schemed_original_table_name(cls: type[T_]) -> str:
        """Return the original table name with schema.

        :returns: original table name with corresponding schema.
        """
        return f"{cls._table_meta.table_schema}.{cls.original_table_name()}"

    @classmethod
    def _retrieve_field(
        cls: type[T_],
        field_name: str,
    ) -> Field[typing.Any]:
        try:
            return typing.cast(
                Field[typing.Any],
                cls.__dict__[field_name],
            )
        except LookupError as lookup_err:
            lookup_err_msg: typing.Final = (
                f"Table `{cls.__name__}` doesn't have `{field_name}` field"
            )
            raise AttributeError(
                lookup_err_msg,
            ) from lookup_err



from __future__ import annotations

from typing import TYPE_CHECKING, Final

from qaspen.querystring.querystring import QueryString
from qaspen.statements.combinable_statements.combinations import (
    CombinableExpression,
)
from qaspen.statements.statement import BaseStatement

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.qaspen_types import FromTable
    from qaspen.statements.select_statement import SelectStatement


class Intersect(CombinableExpression):
    """Class represents `Intersect` object.

    It can contain 1 or more `Intersect` objects.
    """

    def __init__(
        self: Self,
        left_expression: SelectStatement[FromTable] | Intersect,
        right_expression: SelectStatement[FromTable],
    ) -> None:
        self.left_expression: Final = left_expression
        self.right_expression: Final = right_expression

    def querystring(self: Self) -> QueryString:
        """Build `QueryString`."""
        return QueryString(
            self.left_expression.querystring(),
            self.right_expression.querystring(),
            sql_template="{} " + "INTERSECT" + " {}",
        )


class IntersectStatement(BaseStatement):
    """Intersect statement to aggregate `SelectStatement`.

    `intersect_statement` Intersect .
    """

    intersect_statement: Intersect

    def __init__(
        self: Self,
        left_expression: SelectStatement[FromTable] | Intersect,
        right_expression: SelectStatement[FromTable],
    ) -> None:
        self.intersect_statement: Intersect = Intersect(
            left_expression=left_expression,
            right_expression=right_expression,
        )

    def intersect(
        self: Self,
        select_statement: SelectStatement[FromTable],
    ) -> Self:
        """Create `Intersect` with `SelectStatement`.

        ### Parameters:
        - `select_statement`: initialized `SelectStatement`.

        ### Returns:
        `self`.
        """
        self.intersect_statement = Intersect(
            left_expression=self.intersect_statement,
            right_expression=select_statement,
        )
        return self

    def querystring(self: Self) -> QueryString:
        """Build `QueryString`."""
        return self.intersect_statement.querystring()



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

from qaspen.querystring.querystring import QueryString
from qaspen.statements.statement import BaseStatement

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclasses.dataclass
class LimitStatement(BaseStatement):
    """Limit statement for high-level statements.

    It is used in Select/Update/Insert/Delete Statements.

    `limit_number` limit number.
    """

    limit_number: int | None = None

    def limit(
        self: Self,
        limit_number: int,
    ) -> None:
        """Set limit number.

        ### Parameters:
        - `limit_number`: number of the limit.
        """
        self.limit_number = limit_number

    def querystring(self: Self) -> QueryString:
        """Build `QueryString`."""
        if not self.limit_number:
            return QueryString.empty()
        return QueryString(
            self.limit_number,
            sql_template="LIMIT {}",
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Final, Generic

from qaspen.abc.abc_types import DBConnection, Engine

if TYPE_CHECKING:
    import types

    from typing_extensions import Self


class BaseTransaction(ABC, Generic[Engine, DBConnection]):
    """Base class for all possible database transactions."""

    def __init__(
        self: Self,
        engine: Engine,
    ) -> None:
        self.engine: Final = engine

    @abstractmethod
    async def __aenter__(self: Self) -> Self:
        """Enter in the async context manager.

        This method must setup new transaction.

        ### Returns:
        New transaction context manager.
        """

    @abstractmethod
    async def __aexit__(
        self: Self,
        exception_type: type[BaseException] | None,
        exception: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        """Close async context manager.

        Must rollback transaction changes if there is an error.

        ### Parameters:
        - `exception_type`: type of the exception.
        - `exception`: instance of the exception.
        - `traceback`: traceback of the exception.
        """

    @abstractmethod
    async def retrieve_connection(self: Self) -> DBConnection:
        """Retrieve new connection.

        Retrieve new connection from the engine.

        ### Returns:
        Connection from the driver.
        """

    @abstractmethod
    async def begin(self: Self) -> None:
        """Start the transaction."""

    @abstractmethod
    async def commit(self: Self) -> None:
        """Commit the transaction."""

    @abstractmethod
    async def rollback(self: Self) -> None:
        """Rollback the transaction."""



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, Final

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.qaspen_types import MSGSpecStruct, PydanticModel


class RawStatementResult(abc.ABC):
    """Base class for all result statements.

    It will be used when user perform any query
    that can return the result.
    """

    def __init__(
        self: Self,
        engine_result: list[dict[str, Any]],
    ) -> None:
        """Initialize result statement.

        ### Params:
        - `engine_result`: result from the engine.
        """
        self._engine_result: Final = engine_result

    @abc.abstractmethod
    def result(
        self: Self,
    ) -> list[dict[str, Any]]:
        """Return result as-is from engine.

        As any `Engine` must return result
        as `list[dict[str, Any]]`, this method
        must return just clear engine result.
        """


class PydanticStatementResult(abc.ABC):
    """Result as a pydantic model."""

    @abc.abstractmethod
    def to_pydantic(
        self: Self,
        pydantic_model: type[PydanticModel],
    ) -> list[PydanticModel]:
        """Return result as a list of pydantic models.

        You can pass the pydantic mode type in the method and
        result will be transformed into you pydantic model.

        ### Parameters:
        - `pydantic_model`: pydantic model for engine result.

        ### Returns:
        list of `pydantic_model`.
        """


class MSGSpecStatementResult(abc.ABC):
    """Result as a `msgspec` model."""

    @abc.abstractmethod
    def to_msgspec(
        self: Self,
        msgspec_struct: type[MSGSpecStruct],
    ) -> list[MSGSpecStruct]:
        """Return result as a list of msgspec structs.

        You can pass the msgspec struct type in the method and
        result will be transformed into you pydantic model.

        ### Parameters:
        - `msgspec_struct`: msgspec struct for engine result.

        ### Returns:
        list of `msgspec_struct`
        """



from typing import TypeVar

Engine = TypeVar(
    "Engine",
)

DBConnection = TypeVar(
    "DBConnection",
)

EngineConnectionPool = TypeVar(
    "EngineConnectionPool",
)

EngineTransaction = TypeVar(
    "EngineTransaction",
)



import abc
from typing import runtime_checkable

from typing_extensions import Protocol, Self

from qaspen.querystring.querystring import QueryString


@runtime_checkable
class SQLSelectable(Protocol):
    """Protocol for any object that can be used in SQL query."""

    @abc.abstractmethod
    def querystring(self: Self) -> QueryString:
        """Create new QueryString.

        QueryString is the main SQL query building class.
        It can concatenate with different delimiters.

        ### Returns
        `QueryString` or it's subclass.
        """
        ...



import typing


def transform_value_to_sql(value_to_convert: typing.Any) -> str:
    """Convert python value to SQL string.

    ### Parameters:
    - `value_to_convert`: value to convert into SQL string.

    ### Returns:
    SQL string.
    """
    if value_to_convert is None:
        return "NULL"
    if isinstance(value_to_convert, str):
        return f"'{value_to_convert}'"
    if isinstance(value_to_convert, int):
        return f"{value_to_convert}"

    return str(value_to_convert)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Final

from qaspen.querystring.querystring import OrderByQueryString

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.fields.base import Field


class OrderBy:
    """Main class for PostgreSQL OrderBy."""

    def __init__(
        self: Self,
        field: Field[Any],
        ascending: bool = True,
        nulls_first: bool = True,
    ) -> None:
        self.field: Final[Field[Any]] = field  # type: ignore[arg-type]
        self.ascending: Final[bool] = ascending
        self.nulls_first: Final[bool] = nulls_first

    def querystring(self: Self) -> OrderByQueryString:
        """Build `QueryString`."""
        querystring_template: Final[str] = "{} {} {}"
        querystring_arguments: list[str] = [self.field.field_name]

        if self.ascending is True:
            querystring_arguments.append("ASC")
        elif self.ascending is False:
            querystring_arguments.append("DESC")

        if self.nulls_first is True:
            querystring_arguments.append("NULLS FIRST")
        elif self.nulls_first is False:
            querystring_arguments.append("NULLS LAST")

        return OrderByQueryString(
            *querystring_arguments,
            sql_template=querystring_template,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Final

from qaspen.aggregate_functions.base import AggFunction
from qaspen.utils.fields_utils import transform_value_to_sql

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.base.sql_base import SQLSelectable
    from qaspen.clauses.order_by import OrderBy


class Count(AggFunction):
    """Count function.

    The `COUNT()` function is an aggregate function
    that allows you to get the number of rows
    that match a specific condition of a query.
    """

    function_name = "COUNT"

    def __init__(
        self: Self,
        func_argument: SQLSelectable | Any,
        alias: str | None = None,
    ) -> None:
        """Create `COUNT` function.

        ### Parameters:
        - `func_argument`: It's an object with `querystring()` method
        (Field, for example), or any base python class
        (like str, int, etc. and their subclasses).
        - `alias`: name for a `AS` clause in statement.
        """
        super().__init__(
            func_argument,
            alias=alias,
        )


class Coalesce(AggFunction):
    """Coalesce function.

    The COALESCE function accepts an unlimited number of arguments.
    It returns the first argument that is not null.
    If all arguments are null, the COALESCE function will return null.
    """

    function_name = "COALESCE"

    def __init__(
        self: Self,
        *func_argument: SQLSelectable | Any,
        alias: str | None = None,
    ) -> None:
        """Initialize Coalesce function.

        ### Parameters:
        - `func_argument`: arguments for the aggregate function.
            There is can be unlimited number of function arguments.
        - `alias`: alias for the function result.
        """
        super().__init__(
            func_argument,
            alias=alias,
        )


class Agg(AggFunction):
    """Count function.

    The `COUNT()` function is an aggregate function
    that allows you to get the number of rows
    that match a specific condition of a query.
    """

    function_name = "COUNT"

    def __init__(
        self: Self,
        func_argument: SQLSelectable | Any,
        alias: str | None = None,
    ) -> None:
        """Create `AGG` function.

        ### Parameters:
        - `func_argument`: It's an object with `querystring()` method
        (Field, for example), or any base python class
        (like str, int, etc. and their subclasses).
        - `alias`: name for a `AS` clause in statement.
        """
        super().__init__(
            func_argument,
            alias=alias,
        )


class ArrayAgg(AggFunction):
    """ARRAY_AGG function.

    The PostgreSQL ARRAY_AGG() function is an aggregate function
    that accepts a set of values and returns an array
    in which each value in the set is assigned to an element of the array.
    """

    function_name = "ARRAY_AGG"

    def __init__(
        self: Self,
        func_argument: SQLSelectable,
        alias: str | None = None,
        order_by: list[SQLSelectable] | None = None,
        order_by_objs: list[OrderBy] | None = None,
    ) -> None:
        """Create `ARRAY_AGG` function.

        ### Parameters:
        - `func_argument`: arguments for the aggregate function.
            Usually it is Field instance.
        - `alias`: alias for the function result.
        - `order_by`: list of `Field` to order by.
        - `order_by_objs`: list of `OrderBy` objects.
            It can be useful because sometimes you want to add
            `DESC`/`ASC` and `NULLS FIRST/LAST`.
        """
        super().__init__(
            func_argument,
            alias=alias,
        )

        self.order_by: Final = order_by
        self.order_by_objs: Final = order_by_objs

    @property
    def _template_args(self: Self) -> str:
        order_by_args = ""
        order_by_objects_args = ""

        if self.order_by:
            order_by_args = ", ".join(
                ["{}" for _ in self.order_by],
            )
        if self.order_by_objs:
            order_by_objects_args = ", ".join(
                ["{}" for _ in self.order_by_objs],
            )

        if order_by_args and order_by_objects_args:
            return (
                "{} ORDER BY " + f"{order_by_args}, " + order_by_objects_args
            )
        if order_by_args:
            return "{} ORDER BY " + order_by_args
        if order_by_objects_args:
            return "{} ORDER BY " + order_by_objects_args

        return "{}"

    @property
    def _querystring_args(self: Self) -> list[str]:
        querystring_args = super()._querystring_args

        if self.order_by:
            for single_order_by in self.order_by:
                querystring_args.append(
                    str(single_order_by.querystring()),
                )
        if self.order_by_objs:
            for order_by_obj in self.order_by_objs:
                querystring_args.append(
                    str(order_by_obj.querystring()),
                )

        return querystring_args


class Sum(AggFunction):
    """`SUM` function.

    The PostgreSQL SUM() is an aggregate function
    that returns the sum of values or distinct values.
    """

    function_name = "SUM"

    def __init__(
        self: Self,
        func_argument: SQLSelectable,
        alias: str | None = None,
    ) -> None:
        """Create `SUM` function.

        ### Parameters:
        - `func_argument`: It's an object with `querystring()` method
        (Field, for example)
        - `alias`: name for a `AS` clause in statement.
        """
        super().__init__(
            func_argument,
            alias=alias,
        )


class StringAgg(AggFunction):
    """`STRING_AGG` function.

    The PostgreSQL STRING_AGG() function is an aggregate function
    that concatenates a list of strings and places a separator between them.
    The function does not add the separator at the end of the string.
    """

    function_name = "STRING_AGG"

    def __init__(
        self: Self,
        func_argument: SQLSelectable,
        separator: str,
        alias: str | None = None,
        order_by: list[SQLSelectable] | None = None,
        order_by_objs: list[OrderBy] | None = None,
    ) -> None:
        """Create `STRING_AGG` function.

        ### Parameters:
        - `func_argument`: arguments for the aggregate function.
            Tt is Field instance usually.
        - `separator`: separator for strings.
        - `alias`: alias for the function result.
        - `order_by`: list of `Field` to order by.
        - `order_by_objs`: list of `OrderBy` objects.
            It can be useful because sometimes you want to add
            `DESC`/`ASC` and `NULLS FIRST/LAST`.
        """
        super().__init__(
            func_argument,
            alias=alias,
        )

        self.order_by: Final = order_by
        self.order_by_objs: Final = order_by_objs
        self.separator: Final = transform_value_to_sql(separator)

    @property
    def _template_args(self: Self) -> str:
        order_by_args = ""
        order_by_objects_args = ""

        if self.order_by:
            order_by_args = ", ".join(
                ["{}" for _ in self.order_by],
            )
        if self.order_by_objs:
            order_by_objects_args = ", ".join(
                ["{}" for _ in self.order_by_objs],
            )

        if order_by_args and order_by_objects_args:
            return (
                "{}, {} ORDER BY "  # noqa: ISC003
                + f"{order_by_args}, "
                + order_by_objects_args
            )
        if order_by_args:
            return "{}, {} ORDER BY " + order_by_args
        if order_by_objects_args:
            return "{}, {} ORDER BY " + order_by_objects_args

        return "{}, {}"

    @property
    def _querystring_args(self: Self) -> list[str]:
        querystring_args = super()._querystring_args
        querystring_args.append(self.separator)

        if self.order_by:
            for single_order_by in self.order_by:
                querystring_args.append(
                    str(single_order_by.querystring()),
                )
        if self.order_by_objs:
            for order_by_obj in self.order_by_objs:
                querystring_args.append(
                    str(order_by_obj.querystring()),
                )

        return querystring_args


class Max(AggFunction):
    """`MAX` function.

    PostgreSQL MAX function is an aggregate function
    that returns the maximum value in a set of values.
    """

    function_name = "MAX"

    def __init__(
        self: Self,
        func_argument: SQLSelectable,
        alias: str | None = None,
    ) -> None:
        """Create `MAX` function.

        ### Parameters:
        - `func_argument`: It's an object with `querystring()` method
        (Field, for example)
        - `alias`: name for a `AS` clause in statement.
        """
        super().__init__(
            func_argument,
            alias=alias,
        )


class Min(AggFunction):
    """`MIN` function.

    PostgreSQL MIN() function an aggregate function
    that returns the minimum value in a set of values.
    """

    function_name = "MIN"

    def __init__(
        self: Self,
        func_argument: SQLSelectable,
        alias: str | None = None,
    ) -> None:
        """Create `MIN` function.

        ### Parameters:
        - `func_argument`: It's an object with `querystring()` method
        (Field, for example)
        - `alias`: name for a `AS` clause in statement.
        """
        super().__init__(
            func_argument,
            alias=alias,
        )


class Greatest(AggFunction):
    """`GREATEST` function.

    In PostgreSQL, the GREATEST() function returns the largest
    value from the specified values.
    """

    function_name = "GREATEST"

    def __init__(
        self: Self,
        *func_argument: SQLSelectable,
        alias: str | None = None,
    ) -> None:
        """Create `GREATEST` function.

        ### Parameters:
        - `func_argument`: It's an object with `querystring()` method
        (Field, for example)
        - `alias`: name for a `AS` clause in statement.
        """
        super().__init__(
            *func_argument,
            alias=alias,
        )


class Least(AggFunction):
    """`LEAST` function.

    In PostgreSQL, the LEAST() functions returns the smallest
    values from specified values.
    """

    function_name = "LEAST"

    def __init__(
        self: Self,
        *func_argument: SQLSelectable,
        alias: str | None = None,
    ) -> None:
        """Create `LEAST` function.

        ### Parameters:
        - `func_argument`: It's an object with `querystring()` method
        (Field, for example)
        - `alias`: name for a `AS` clause in statement.
        """
        super().__init__(
            *func_argument,
            alias=alias,
        )



from __future__ import annotations

import dataclasses
import functools
import operator
from typing import TYPE_CHECKING, Any, Final, Iterable

from qaspen.base.sql_base import SQLSelectable
from qaspen.querystring.querystring import FilterQueryString, QueryString
from qaspen.statements.combinable_statements.combinations import (
    CombinableExpression,
)
from qaspen.statements.statement import BaseStatement
from qaspen.utils.fields_utils import transform_value_to_sql

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.fields.base import Field
    from qaspen.fields.operators import BaseOperator


class EmptyValue:
    """Class represents that value isn't passed.

    It's necessary because `None` is valid value.
    """


EMPTY_VALUE = EmptyValue()


class Filter(CombinableExpression):
    """Class that represents any Filter in PostgreSQL.

    It is used in the places where you use comparisons like
    `eq` and etc.
    This is usually used for `WHERE` and `ON` clauses.
    """

    def __init__(
        self: Self,
        field: Field[Any],
        operator: type[BaseOperator],
        comparison_value: EmptyValue | Field[Any] | Any = EMPTY_VALUE,
        comparison_values: EmptyValue | Iterable[Any] = EMPTY_VALUE,
    ) -> None:
        self.field: Field[Any] = field  # type: ignore[arg-type]
        self.operator: type[BaseOperator] = operator

        self.comparison_value: EmptyValue | Any = comparison_value
        self.comparison_values: Final = comparison_values

    def querystring(self: Self) -> FilterQueryString:
        """Build new `FilterQueryString`."""
        compare_value: str = ""
        if self.comparison_value is not EMPTY_VALUE:
            if isinstance(self.comparison_value, SQLSelectable):
                compare_value = str(self.comparison_value.querystring())
            else:
                compare_value = transform_value_to_sql(self.comparison_value)
        elif self.comparison_values is not EMPTY_VALUE:
            compare_value = ", ".join(
                [
                    transform_value_to_sql(comparison_value)
                    for comparison_value in self.comparison_values  # type: ignore[union-attr]
                ],
            )

        return FilterQueryString(
            self.field.field_name,
            compare_value,
            sql_template=self.operator.operation_template,
        )


class FilterBetween(CombinableExpression):
    """Class that represents BETWEEN filter in PostgreSQL.

    It is used in the places where you use comparisons like
    `eq` and etc.
    This is usually used for `WHERE` and `ON` clauses.
    """

    def __init__(
        self: Self,
        field: Field[Any],
        operator: type[BaseOperator],
        left_comparison_value: Any,
        right_comparison_value: Any,
    ) -> None:
        self.field: Field[Any] = field  # type: ignore[arg-type]
        self.operator: type[BaseOperator] = operator

        self.left_comparison_value: Any = left_comparison_value
        self.right_comparison_value: Any = right_comparison_value

    def querystring(self: Self) -> FilterQueryString:
        """Build new `FilterQueryString`."""
        from qaspen.fields.base import BaseField

        left_value: str = (
            self.left_comparison_value.field_name
            if isinstance(self.left_comparison_value, BaseField)
            else transform_value_to_sql(self.left_comparison_value)
        )

        right_value: str = (
            self.right_comparison_value.field_name
            if isinstance(self.right_comparison_value, BaseField)
            else transform_value_to_sql(self.right_comparison_value)
        )

        return FilterQueryString(
            self.field.field_name,
            left_value,
            right_value,
            sql_template=self.operator.operation_template,
        )


class FilterExclusive(CombinableExpression):
    """Special class that can isolate Filters in brackets."""

    def __init__(
        self: Self,
        comparison: CombinableExpression,
    ) -> None:
        """Initialize FilterExclusive.

        This class isolates expressions, so you can build
        more complex queries.
        """
        self.comparison: CombinableExpression = comparison

    def querystring(self: Self) -> FilterQueryString:
        """Build new `FilterQueryString`."""
        return FilterQueryString(
            *self.comparison.querystring().template_arguments,
            sql_template=(
                "(" + self.comparison.querystring().sql_template + ")"
            ),
        )


@dataclasses.dataclass
class FilterStatement(BaseStatement):
    """Filter statement for high-level statements.

    It is used in Select/Update/Insert/Delete Statements.
    """

    filter_expressions: list[CombinableExpression] = dataclasses.field(
        default_factory=list,
    )

    def add_filter(
        self: Self,
        *filter_argument: CombinableExpression,
    ) -> Self:
        """Add new CombinableExpression's.

        ### Parameters:
        - `filter_argument`: CombinableExpressions.

        ### Returns:
        `self`.
        """
        self.filter_expressions.extend(
            filter_argument,
        )
        return self

    def querystring(self: Self) -> QueryString:
        """Build new `QueryString`.

        Adds all statements to each other.

        If there is no `filter_expressions`, then
        return `EmptyQueryString`.

        ### Returns:
        `QueryString`
        """
        if not self.filter_expressions:
            return QueryString.empty()
        final_where: QueryString = functools.reduce(
            operator.add,
            [
                FilterQueryString(
                    *filter_expression.querystring().template_arguments,
                    sql_template=filter_expression.querystring().sql_template,
                )
                for filter_expression in self.filter_expressions
            ],
        )

        return QueryString(
            *final_where.template_arguments,
            sql_template=f"WHERE {final_where.sql_template}",
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Final, Generator, Generic, List

from qaspen.base.sql_base import SQLSelectable
from qaspen.qaspen_types import FromTable
from qaspen.querystring.querystring import QueryString
from qaspen.statements.base import Executable
from qaspen.statements.combinable_statements.combinations import (
    CombinableExpression,
)
from qaspen.statements.statement import BaseStatement

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.abc.db_engine import BaseEngine
    from qaspen.statements.select_statement import SelectStatement


class SQLUnion(CombinableExpression):
    """Main class for PostgreSQL `UNION`."""

    def __init__(
        self: Self,
        left_expression: SelectStatement[FromTable] | SQLUnion,
        right_expression: SelectStatement[FromTable],
        union_all: bool = False,
    ) -> None:
        self.left_expression: Final = left_expression
        self.right_expression: Final = right_expression
        self.union_all: bool = union_all

    def querystring(self: Self) -> QueryString:
        """Build `QueryString`."""
        union_operator: str = "UNION ALL" if self.union_all else "UNION"
        return QueryString(
            self.left_expression.querystring(),
            self.right_expression.querystring(),
            sql_template="{} " + union_operator + " {}",
        )


class UnionStatement(
    BaseStatement,
    SQLSelectable,
    Executable[List[Dict[str, Any]]],
    Generic[FromTable],
):
    """Union statement.

    Create with two or more `SelectStatements`.

    `union_statement` main SQLUnion object.
    """

    union_statement: SQLUnion

    def __init__(
        self: Self,
        left_expression: SelectStatement[FromTable] | SQLUnion,
        right_expression: SelectStatement[FromTable],
        union_all: bool = False,
    ) -> None:
        self.union_statement: SQLUnion = SQLUnion(
            left_expression=left_expression,
            right_expression=right_expression,
            union_all=union_all,
        )

    def __await__(
        self: Self,
    ) -> Generator[None, None, list[dict[str, Any]]]:
        """SelectStatement can be awaited.

        Example:
        -------
        ```
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()

        async def main() -> None:
            list_of_buns = await Buns.select()
            print(list_of_buns)
        ```
        """
        first_select_statement: Final = self._start_select_statement(
            statement=self.union_statement.left_expression,  # type: ignore[arg-type]
        )

        engine: BaseEngine[
            Any,
            Any,
            Any,
        ] | None = (
            first_select_statement._from_table._table_meta.database_engine
        )
        if not engine:
            engine_err_msg: Final = "There is no database engine."
            raise AttributeError(engine_err_msg)

        return self.execute(engine=engine).__await__()

    async def execute(
        self: Self,
        engine: BaseEngine[Any, Any, Any],
    ) -> list[dict[str, Any]]:
        """Execute SQL query and return result."""
        raw_query_result: list[dict[str, Any]] = await engine.execute(
            querystring=self.querystring(),
        )
        return raw_query_result

    def union(
        self: Self,
        select_statement: SelectStatement[FromTable],
        union_all: bool = False,
    ) -> Self:
        """Add new `SelectStatement` to the `SQLUnion`.

        ### Parameters:
        - `select_statement`: initialized `SelectStatement`.
        - `union_all`: use `UNION` or `UNION ALL`.

        ### Returns:
        `self`
        """
        self.union_statement = SQLUnion(
            left_expression=self.union_statement,
            right_expression=select_statement,
            union_all=union_all,
        )
        return self

    def querystring(self: Self) -> QueryString:
        """Build new `QueryString`."""
        return self.union_statement.querystring()

    def _start_select_statement(
        self: Self,
        statement: SelectStatement[FromTable] | SQLUnion,
    ) -> SelectStatement[FromTable]:
        """Retrieve first SelectStatement in the UnionStatement.

        ### Parameters
        :param statement: Union or SelectStatement.

        ### Returns
        :returns: SelectStatement.
        """
        if isinstance(statement, SQLUnion):
            return self._start_select_statement(
                statement=statement.left_expression,  # type: ignore[arg-type]
            )
        return statement



import typing

from msgspec import Struct
from pydantic import BaseModel

if typing.TYPE_CHECKING:
    from qaspen.table.base_table import BaseTable


FromTable = typing.TypeVar(
    "FromTable",
    bound="BaseTable",
)

FieldType = typing.TypeVar(
    "FieldType",
)

FieldDefaultType = typing.Union[
    FieldType,
    typing.Callable[
        [],
        FieldType,
    ],
    None,
]

CallableDefaultType = typing.Callable[
    [],
    FieldType,
]

PydanticModel = typing.TypeVar(
    "PydanticModel",
    bound=BaseModel,
)

MSGSpecStruct = typing.TypeVar(
    "MSGSpecStruct",
    bound=Struct,
)



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, Generator, Generic, TypeVar

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.abc.db_engine import BaseEngine


StatementResultType = TypeVar(
    "StatementResultType",
)


class Executable(abc.ABC, Generic[StatementResultType]):
    """Show that statement can be executed."""

    @abc.abstractmethod
    def __await__(
        self: Self,
    ) -> Generator[None, None, StatementResultType]:
        """Make statement awaitable."""

    @abc.abstractmethod
    async def execute(
        self: Self,
        engine: BaseEngine[Any, Any, Any],
    ) -> StatementResultType:
        """Execute SQL query and return result."""



import datetime

from qaspen.sql_type.base import SQLType


class SmallInt(SQLType):
    """Represent `SMALLINT` PostgreSQL type in the python."""

    python_type = int


class Integer(SQLType):
    """Represent `INTEGER` PostgreSQL type in the python."""

    python_type = int


class BigInt(SQLType):
    """Represent `BigInt` PostgreSQL type in the python."""

    python_type = int


class Numeric(SQLType):
    """Represent `Numeric` PostgreSQL type in the python."""

    python_type = int


class Decimal(SQLType):
    """Represent `Decimal` PostgreSQL type in the python."""

    python_type = int


class Real(SQLType):
    """Represent `REAL` PostgreSQL type in the python."""

    python_type = (float, str)


class DoublePrecision(SQLType):
    """Represent `DOUBLEPRECISION` PostgreSQL type in the python."""

    python_type = (float, str)


class Boolean(SQLType):
    """Represent `BOOLEAN` PostgreSQL type in the python."""

    python_type = bool


class VarChar(SQLType):
    """Represent `VARCHAR` PostgreSQL type in the python."""

    python_type = str


class Text(SQLType):
    """Represent `TEXT` PostgreSQL type in the python."""

    python_type = str


class Char(SQLType):
    """Represent `CHAR` PostgreSQL type in the python."""

    python_type = str


class Date(SQLType):
    """Represent `DATE` PostgreSQL type in the python."""

    python_type = datetime.date


class Time(SQLType):
    """Represent `TIME` PostgreSQL type in the python."""

    python_type = datetime.time


class Timestamp(SQLType):
    """Represent `TIMESTAMP` PostgreSQL type in the python."""

    python_type = datetime.datetime


class Interval(SQLType):
    """Represent `INTERVAL` PostgreSQL type in the python."""

    python_type = datetime.timedelta



from __future__ import annotations

import abc
import copy
import dataclasses
import types
from typing import TYPE_CHECKING, Any, Callable, Final, Generic, Union, cast

from qaspen.base.operators import AllOperator, AnyOperator
from qaspen.exceptions import (
    FieldComparisonError,
    FieldDeclarationError,
    FieldValueValidationError,
    FilterComparisonError,
)
from qaspen.fields import operators
from qaspen.qaspen_types import (
    CallableDefaultType,
    FieldDefaultType,
    FieldType,
)
from qaspen.querystring.querystring import QueryString
from qaspen.statements.combinable_statements.filter_statement import (
    Filter,
    FilterBetween,
)

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.base.sql_base import SQLSelectable
    from qaspen.sql_type.base import SQLType
    from qaspen.table.base_table import BaseTable


class EmptyFieldValue:
    """Indicates that field wasn't queried from the database."""

    def __str__(self: Self) -> str:
        return self.__class__.__name__


EMPTY_FIELD_VALUE = EmptyFieldValue()


@dataclasses.dataclass
class FieldData(Generic[FieldType]):
    """All field data.

    ### Fields
    `field_name` - Name of the field

    `from_table` - From what table field is.

    `is_null` - is field can be NULL.

    `field_value` - value of the field in the database or
    if you create table instance with fields manually.

    `default` - default value for the field.
    If it is callable, do not use it in the database,
    use it in python and than save in the database.

    `prefix` - prefix for the field in the query.

    `alias` - alias of the field.

    `in_join` - mark that field is used in join.
    """

    field_name: str
    from_table: type[BaseTable] = None  # type: ignore[assignment]
    is_null: bool = False
    field_value: FieldType | EmptyFieldValue | None = EMPTY_FIELD_VALUE
    default: str | None = None
    callable_default: Callable[[], FieldType] | None = None
    prefix: str = ""
    alias: str = ""
    in_join: bool = False


class BaseField(Generic[FieldType], abc.ABC):
    """Base field class for all Fields."""

    _field_data: FieldData[FieldType]
    _sql_type: type[SQLType]

    def __set_name__(
        self: Self,
        owner: Any,
        field_name: str,
    ) -> None:
        """Set name for the field.

        field_name is equal to the name of the Field variable.

        Example:
        -------
        ```
        class MyTable(BaseTable):
            # name of the field in the database is `meme_lord`
            meme_lord: VarCharField = VarCharField()
        ```
        """
        self._field_data.from_table = owner
        self._field_data.field_name = field_name

    @abc.abstractmethod
    def _make_field_create_statement(
        self: Self,
    ) -> str:
        ...

    @abc.abstractmethod
    def _validate_field_value(
        self: Self,
        field_value: FieldType | None,
    ) -> None:
        """Validate field value.

        It must raise an error if something goes wrong
        or not return anything.
        """
        ...

    @abc.abstractmethod
    def _validate_default_value(
        self: Self,
        default_value: FieldType | None,
    ) -> None:
        """Validate field default value.

        It must raise an error in validation failed.
        """
        ...

    @abc.abstractmethod
    def _prepare_default_value(
        self: Self,
        default_value: FieldType,
    ) -> str | None:
        """Prepare default value to specify it in Field declaration.

        It uses only in the method `_field_default`.

        :returns: prepared default value.
        """

    @property
    def value(self: Self) -> FieldType | EmptyFieldValue | None:
        """Return value of the field.

        It's valid only for fields after executing statement,
        otherwise it returns `None`.

        If you don't select field in the `.select()` and
        try to get its value after `as_objects()` method,
        you will retrieve `EmptyFieldValue` because it indicates
        that you don't select this Field.

        ### Returns:
        `Python type of the field` or `EmptyFieldValue` or `None`.
        """
        return self._field_data.field_value

    @property
    def _table_name(self: Self) -> str:
        """Return table name of this field.

        It's only the original table name, not aliased.

        ### Return
        `str` table name.
        """
        return self._field_data.from_table.original_table_name()

    @property
    def _schemed_table_name(self: Self) -> str:
        """Return aliased table name with schema of this field.

        ### Return
        `str` table name.
        """
        return self._field_data.from_table.schemed_table_name()

    @property
    def alias(self: Self) -> str | None:
        """Return alias to the field if it exists.

        ### Returns:
        `alias` or `None`.
        """
        return self._field_data.alias

    @property
    def field_name(self: Self) -> str:
        """Return field name with prefix and alias.

        ### Prefix logic:
        If `from_table` has alias than use this alias.

        If this field has a prefix than use prefix.

        Else use original name of the field's Table.

        ### Alias logic:
        If this Field has an alias, use it.

        ### Return
        `Field` as a `str`.
        """
        prefix: str = (
            self._field_data.from_table._table_meta.alias
            or self._field_data.prefix
            or self._field_data.from_table.table_name()
        )
        field_name: str = f"{prefix}.{self._field_data.field_name}"
        if alias := self._field_data.alias:
            field_name += f" AS {alias}"

        return field_name

    @property
    def _original_field_name(self: Self) -> str:
        """Return name of the field without prefix and alias.

        ### Return
        `str` Field name.
        """
        return self._field_data.field_name

    @property
    def _default(self: Self) -> str | None:
        """Return default value of the field.

        This default is already converted into SQL string.
        Or None.

        ### Return
        default value.
        """
        return self._field_data.default

    @property
    def _callable_default(
        self: Self,
    ) -> CallableDefaultType[FieldType] | None:
        """Return callable object for default value for the field.

        This default value will be called in the statements
        that can be create new raws in the database.

        """
        return self._field_data.callable_default

    @property
    def _is_null(self: Self) -> bool:
        """Return flag that field can be `NULL`.

        ### Return
        `bool`
        """
        return self._field_data.is_null

    @property
    def _field_null(self: Self) -> str:
        """Return `NOT NULL` string if field cannot be NULL.

        ### Return
        `str`.
        """
        return "NOT NULL" if not self._is_null else ""

    @property
    def _field_default(self: Self) -> str:
        if self._default:
            return f"DEFAULT {self._default}" if self._default else ""
        return ""

    @property
    def _field_type(self: Self) -> str:
        """Property for final SQL field Type.

        It can be changed in subclasses.

        ### Return
        SQL `string`.
        """
        return str(self._sql_type.querystring())


if TYPE_CHECKING:
    from qaspen.table.base_table import BaseTable

OperatorTypes = Union[AnyOperator, AllOperator]


class Field(BaseField[FieldType]):
    """Main Field class.

    All subclasses must be inherited from this class.

    `_available_comparison_types` defines what types
    can be used in comparison methods, like `__eq__`.

    `_set_available_types` defines what types can be
    set as a value to the field.
    """

    _available_comparison_types: tuple[type, ...]
    _set_available_types: tuple[type, ...]

    def __init__(
        self: Self,
        *args: Any,
        is_null: bool = False,
        default: FieldDefaultType[FieldType] = None,
        db_field_name: str | None = None,
    ) -> None:
        """Create new Field instance.

        It's not possible to use not keyword arguments for
        specifying parameters.

        ### Parameters:
        - `is_null`: Defines is Field can be `NULL` or not.
        - `default`: Default value for the Field.
            This default value can be callable object.
            Note! This value will be set at the python level.
        - `db_field_name`: name of the field in the database.
        """
        if args:
            args_err_msg: Final = "Use only keyword arguments."
            raise FieldDeclarationError(args_err_msg)

        if is_null and default:
            err_msg: Final = (
                "It's not possible to specify is_null and default. "
                "Specify either is_null or default"
            )
            raise FieldDeclarationError(err_msg)

        self._validate_default_value(
            default_value=default,
        )

        default_value: str | None = None
        callable_default_value: CallableDefaultType[FieldType] | None = None

        if callable(default):
            callable_default_value = default
        elif default is not None:
            default_value = self._prepare_default_value(
                default_value=default,
            )

        self._field_data: FieldData[FieldType] = FieldData(
            field_name=db_field_name or "",
            is_null=is_null,
            default=default_value,
            callable_default=callable_default_value,
        )

    def __get__(
        self: Self,
        instance: BaseTable | None,
        owner: type[BaseTable] | None,
    ) -> Field[FieldType]:
        try:
            return cast(
                Field[FieldType],
                instance.__dict__[self._original_field_name],
            )
        except (AttributeError, KeyError):
            return cast(
                Field[FieldType],
                owner.__dict__[self._original_field_name],
            )

    def __set__(
        self: Self,
        instance: object,
        value: FieldType | EmptyFieldValue | None,
    ) -> None:
        field: Field[FieldType]
        if isinstance(value, EmptyFieldValue) or value is None:
            field = instance.__dict__[self._original_field_name]
            field._field_data.field_value = value
            return

        if isinstance(value, self.__class__):
            instance.__dict__[self._original_field_name] = value
            return

        if not isinstance(value, self._set_available_types):
            err_msg: Final = (
                f"Can't assign not {self._sql_type.querystring()} "
                f"type to {self.__class__.__name__}",
            )
            raise TypeError(err_msg)

        self._validate_field_value(
            field_value=value,  # type: ignore[arg-type]
        )
        field = instance.__dict__[self._original_field_name]
        field._field_data.field_value = value  # type: ignore[assignment]

    def querystring(self: Self) -> QueryString:
        """Build QueryString class.

        ### Returns:
        Built `QueryString`.
        """
        return QueryString(
            self.field_name,
            sql_template="{}",
        )

    def in_(
        self: Self,
        *comparison_values: FieldType,
        subquery: SQLSelectable | None = None,
    ) -> Filter:
        """`IN` PostgreSQL clause.

        It allows you to use `IN` clause.
        You can specify either unlimited number of `comparison` values
        or `subquery`.

        ### Parameters:
        - `comparison_values`: values for `IN` clause,
            they must be correct type. For example, if you are
            working with string Field, you have to use str objects.
        - `subquery`: Any object that provides `queryset()` method.

        ### Returns:
        Initialized `Filter` class.

        Example:
        -------
        ```python
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()


        select_statement = (
            Buns
            .select()
            .where(
                Buns.name.in_(
                    "Awesome",
                    "Qaspen",
                    "Project",
                )
            )
        )

        select_statement = (
            Buns
            .select()
            .where(
                Buns.name.in_(
                    subquery=Buns.select(
                        Buns.name,
                    ),
                ),
            )
        )
        ```
        """
        if subquery and comparison_values:
            args_err_msg: Final = (
                "It's not possible to specify subquery "
                "with positional arguments in `in_` method. "
                "Please choose either subquery or positional arguments.",
            )
            raise FilterComparisonError(args_err_msg)

        for comparison_value in comparison_values:
            is_valid_type: bool = isinstance(
                comparison_value,
                self._available_comparison_types,
            )
            if not is_valid_type:
                err_msg = (
                    f"It's impossible to use `IN` operator "
                    f"to compare {self.__class__.__name__} "
                    f"and {type(comparison_value)}",
                )
                raise FieldComparisonError(err_msg)

        where_parameters: dict[str, Any] = {
            "field": self,
            "operator": operators.InOperator,
        }

        if subquery:
            where_parameters["comparison_value"] = subquery
        elif comparison_values:
            where_parameters["comparison_values"] = comparison_values

        return Filter(**where_parameters)

    def not_in(
        self: Self,
        *comparison_values: FieldType,
        subquery: SQLSelectable | None = None,
    ) -> Filter:
        """`NOT IN` PostgreSQL clause.

        It allows you to use `NOT IN` clause.
        You can specify either unlimited number of `comparison` values
        or `subquery`.

        ### Parameters:
        - `comparison_values`: values for `NOT IN` clause,
            they must be correct type. For example, if you are
            working with string Field, you have to use str objects.
        - `subquery`: Any object that provides `queryset()` method.

        ### Returns:
        Initialized `Filter` class.

        Example:
        -------
        ```python
        class Buns(BaseTable, table_name="buns"):
            name: VarCharField = VarCharField()


        select_statement = (
            Buns
            .select()
            .where(
                Buns.name.not_in(
                    "Awesome",
                    "Qaspen",
                    "Project",
                )
            )
        )

        select_statement = (
            Buns
            .select()
            .where(
                Buns.name.not_in(
                    subquery=Buns.select(
                        Buns.name,
                    ),
                ),
            )
        )
        ```
        """
        if subquery and comparison_values:
            args_err_msg: Final = (
                "It's not possible to specify subquery "
                "with positional arguments in `not_in` method. "
                "Please choose either subquery or positional arguments.",
            )
            raise FilterComparisonError(args_err_msg)

        for comparison_value in comparison_values:
            is_valid_type: bool = isinstance(
                comparison_value,
                self._available_comparison_types,
            )
            if not is_valid_type:
                err_msg = (
                    f"It's impossible to use `NOT IN` operator "
                    f"to compare {self.__class__.__name__} "
                    f"and {type(comparison_value)}",
                )
                raise FieldComparisonError(err_msg)

        where_parameters: dict[str, Any] = {
            "field": self,
            "operator": operators.NotInOperator,
        }

        if subquery:
            where_parameters["comparison_value"] = subquery
        elif comparison_values:
            where_parameters["comparison_values"] = comparison_values

        return Filter(**where_parameters)

    def between(
        self: Self,
        left_value: FieldType | Field[Any],
        right_value: FieldType | Field[Any],
    ) -> FilterBetween:
        """`BETWEEN` PostgreSQL clause.

        It allows you to use `BETWEEN` clause.

        ### Parameters:
        - `left_value`: left-side value in the between clause.
        - `right_value`: right-side value in the between clause.

        ### Returns:
        Initialized `FilterBetween`.
        """
        is_valid_type: Final[bool] = all(
            (
                isinstance(
                    left_value,
                    self._available_comparison_types,
                ),
                isinstance(
                    right_value,
                    self._available_comparison_types,
                ),
            ),
        )
        if is_valid_type:
            return FilterBetween(
                field=self,
                operator=operators.BetweenOperator,
                left_comparison_value=left_value,
                right_comparison_value=right_value,
            )

        type_err_msg = (
            f"Incorrect type of one of the values "
            f"in `BETWEEN operator`. "
            f"You can use one of these - {self._available_comparison_types}",
        )
        raise FieldComparisonError(type_err_msg)

    def __eq__(  # type: ignore[override]
        self: Self,
        comparison_value: FieldType | Field[Any] | OperatorTypes,
    ) -> Filter:
        if comparison_value is None:
            return Filter(
                field=self,
                operator=operators.IsNullOperator,
            )
        if isinstance(comparison_value, self._available_comparison_types):
            return Filter(
                field=self,
                comparison_value=comparison_value,
                operator=operators.EqualOperator,
            )

        comparison_err_msg: Final = (
            f"It's impossible to use `!=` operator "
            f"to compare {self.__class__.__name__} "
            f"and {type(comparison_value)}",
        )
        raise FieldComparisonError(comparison_err_msg)

    def eq(
        self: Self,
        comparison_value: FieldType | Field[Any] | OperatorTypes,
    ) -> Filter:
        """Analog for `==` (`__eq__` method) operation.

        Works exactly the same. It exists just because some
        people prefer to use methods instead of python comparison.

        ### Parameters:
        - `comparison_value`: value to compare with.

        ### Returns:
        Initialized `Filter`.
        """
        return self.__eq__(comparison_value)

    def __ne__(  # type: ignore[override]
        self: Self,
        comparison_value: FieldType | Field[Any] | OperatorTypes,
    ) -> Filter:
        if comparison_value is None:
            return Filter(
                field=self,
                operator=operators.IsNotNullOperator,
            )
        if isinstance(comparison_value, self._available_comparison_types):
            return Filter(
                field=self,
                comparison_value=comparison_value,
                operator=operators.NotEqualOperator,
            )

        comparison_err_msg: Final = (
            f"It's impossible to use `!=` operator "
            f"to compare {self.__class__.__name__} "
            f"and {type(comparison_value)}",
        )
        raise FieldComparisonError(comparison_err_msg)

    def neq(
        self: Self,
        comparison_value: FieldType | OperatorTypes,
    ) -> Filter:
        """Analog for `!=` (`__ne__` method) operation.

        Works exactly the same. It exists just because some
        people prefer to use methods instead of python comparison.

        ### Parameters:
        - `comparison_value`: value to compare with.

        ### Returns:
        Initialized `Filter`.
        """
        return self.__ne__(comparison_value)

    def __gt__(
        self: Self,
        comparison_value: FieldType | OperatorTypes,
    ) -> Filter:
        if isinstance(comparison_value, self._available_comparison_types):
            return Filter(
                field=self,
                comparison_value=comparison_value,
                operator=operators.GreaterOperator,
            )

        comparison_err_msg: Final = (
            f"It's impossible to use `>` operator "
            f"to compare {self.__class__.__name__} "
            f"and {type(comparison_value)}",
        )
        raise FieldComparisonError(comparison_err_msg)

    def gt(
        self: Self,
        comparison_value: FieldType | OperatorTypes,
    ) -> Filter:
        """Analog for `>` (`__gt__` method) operation.

        Works exactly the same. It exists just because some
        people prefer to use methods instead of python comparison.

        ### Parameters:
        - `comparison_value`: value to compare with.

        ### Returns:
        Initialized `Filter`.
        """
        return self.__gt__(comparison_value)

    def __ge__(
        self: Self,
        comparison_value: FieldType | OperatorTypes,
    ) -> Filter:
        if isinstance(comparison_value, self._available_comparison_types):
            return Filter(
                field=self,
                comparison_value=comparison_value,
                operator=operators.GreaterEqualOperator,
            )
        comparison_err_msg: Final = (
            f"It's impossible to use `>=` operator "
            f"to compare {self.__class__.__name__} "
            f"and {type(comparison_value)}",
        )
        raise FieldComparisonError(comparison_err_msg)

    def gte(
        self: Self,
        comparison_value: FieldType | OperatorTypes,
    ) -> Filter:
        """Analog for `>=` (`__ge__` method) operation.

        Works exactly the same. It exists just because some
        people prefer to use methods instead of python comparison.

        ### Parameters:
        - `comparison_value`: value to compare with.

        ### Returns:
        Initialized `Filter`.
        """
        return self.__ge__(comparison_value)

    def __lt__(
        self: Self,
        comparison_value: FieldType | OperatorTypes,
    ) -> Filter:
        if isinstance(comparison_value, self._available_comparison_types):
            return Filter(
                field=self,
                comparison_value=comparison_value,
                operator=operators.LessOperator,
            )
        comparison_err_msg: Final = (
            f"It's impossible to use `<` operator "
            f"to compare {self.__class__.__name__} "
            f"and {type(comparison_value)}",
        )
        raise FieldComparisonError(comparison_err_msg)

    def lt(
        self: Self,
        comparison_value: FieldType | OperatorTypes,
    ) -> Filter:
        """Analog for `<` (`__lt__` method) operation.

        Works exactly the same. It exists just because some
        people prefer to use methods instead of python comparison.

        ### Parameters:
        - `comparison_value`: value to compare with.

        ### Returns:
        Initialized `Filter`.
        """
        return self.__lt__(comparison_value)

    def __le__(
        self: Self,
        comparison_value: FieldType | OperatorTypes,
    ) -> Filter:
        if isinstance(comparison_value, self._available_comparison_types):
            return Filter(
                field=self,
                comparison_value=comparison_value,
                operator=operators.LessEqualOperator,
            )
        comparison_err_msg: Final = (
            f"It's impossible to use `<=` operator "
            f"to compare {self.__class__.__name__} "
            f"and {type(comparison_value)}",
        )
        raise FieldComparisonError(comparison_err_msg)

    def lte(
        self: Self,
        comparison_value: FieldType | OperatorTypes,
    ) -> Filter:
        """Analog for `<=` (`__le__` method) operation.

        Works exactly the same. It exists just because some
        people prefer to use methods instead of python comparison.

        ### Parameters:
        - `comparison_value`: value to compare with.

        ### Returns:
        Initialized `Filter`.
        """
        return self.__le__(comparison_value)

    def with_alias(
        self: Self,
        alias_name: str,
    ) -> Self:
        """Set alias to the field.

        ### Parameters:
        - `alias_name`: name of the alias to the field.

        ### Returns
        `Field` with new prefix.
        """
        return self._with_alias(
            alias=alias_name,
        )

    def _is_the_same_field(
        self: Self,
        other_field: BaseField[FieldType],
    ) -> bool:
        """Compare two fields.

        Return `True` if they are the same, else `False`.
        They are equal if they `_field_data`s are the same.

        ### Parameters:
        - `other_field`: field to compare with.

        ### Returns:
        Are fields the same or not.
        """
        return self._field_data == other_field._field_data

    def _with_prefix(self: Self, prefix: str) -> Field[FieldType]:
        """Give Field a prefix.

        Make a Field deepcopy and set new prefix.

        ### Parameters
        - `prefix`: prefix for the field.

        ### Returns
        `Field` with new prefix.
        """
        field: Field[FieldType] = copy.deepcopy(self)
        field._field_data.prefix = prefix
        return field

    def _with_alias(self: Self, alias: str) -> Self:
        """Give Field an alias.

        Make a Field deepcopy and set new alias.

        ### Parameters
        - `alias`: alias for the field.

        ### Returns
        `Field` with new alias.
        """
        field: Self = copy.deepcopy(self)
        field._field_data.alias = alias
        return field

    def _validate_field_value(
        self: Self,
        field_value: FieldType | None,
    ) -> None:
        """Validate field value.

        If `field_value` is None but field declared as NOT NULL,
        throw an error.

        :param field_value: new value for the field.

        :raises FieldValueValidationError: if the `max_length` is exceeded.
        """
        if field_value is None and self._is_null:
            return

    def _validate_default_value(
        self: Self,
        default_value: FieldDefaultType[FieldType],
    ) -> None:
        if default_value is None or types.FunctionType == type(default_value):
            return

        try:
            self._validate_field_value(
                field_value=default_value,  # type: ignore[arg-type]
            )
        except FieldValueValidationError as exc:
            validation_err_msg: Final = (
                f"Wrong default value in the field "
                f"{self.__class__.__name__}",
            )
            raise FieldValueValidationError(
                validation_err_msg,
            ) from exc

        if not isinstance(default_value, self._set_available_types):
            type_err_msg: Final = (
                f"Wrong default value in the field "
                f"{self.__class__.__name__}",
            )
            raise FieldValueValidationError(type_err_msg)

    def _make_field_create_statement(
        self: Self,
    ) -> str:
        return (
            f"{self._original_field_name} {self._sql_type.sql_type()} "
            f"{self._field_null} {self._field_default}"
        )

    def _prepare_default_value(
        self: Self,
        default_value: FieldType | None,
    ) -> str | None:
        """Prepare default value to specify it in Field declaration.

        It uses only in the method `_field_default`.

        :returns: prepared default value.
        """
        return (
            str(default_value) if default_value is not None else default_value
        )



import abc

from typing_extensions import Self

from qaspen.querystring.querystring import QueryString


class BaseStatement(abc.ABC):
    """Base statement all statements."""

    @abc.abstractmethod
    def querystring(self: Self) -> QueryString:
        """Build `QueryString`."""
        ...



from typing import Final

MAX_STRING_FIELD_LENGTH: Final[int] = 10485760


def validate_max_length(max_length: int) -> bool:
    """Validate length of the string.

    ### Returns:
    boolean flag.
    """
    return max_length >= MAX_STRING_FIELD_LENGTH



from __future__ import annotations

import copy
import dataclasses
from typing import TYPE_CHECKING, Any, ClassVar, Final

from qaspen.fields.base import EmptyFieldValue, Field

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.abc.db_engine import BaseEngine


@dataclasses.dataclass
class MetaTableData:
    """All data about the table."""

    table_name: str = ""
    table_schema: str = "public"
    abstract: bool = False
    table_fields: dict[str, Field[Any]] = dataclasses.field(
        default_factory=dict,
    )
    database_engine: BaseEngine[Any, Any, Any] | None = None
    alias: str | None = None


class MetaTable:
    """Meta Table."""

    _table_meta: MetaTableData = MetaTableData()
    _subclasses: ClassVar[list[type[MetaTable]]] = []

    def __init_subclass__(
        cls: type[MetaTable],
        table_name: str | None = None,
        table_schema: str = "public",
        abstract: bool = False,
        **kwargs: Any,
    ) -> None:
        if not table_name:
            table_name = cls.__name__.lower()

        table_fields: Final[dict[str, Field[Any]],] = cls._parse_table_fields()

        cls._table_meta = MetaTableData(
            table_name=table_name,
            table_schema=table_schema,
            abstract=abstract,
            table_fields=table_fields,
        )

        cls._subclasses.append(cls)

        super().__init_subclass__(**kwargs)

    def __init__(self: Self, **fields_values: Any) -> None:
        for table_field in self._table_meta.table_fields.values():
            setattr(
                self,
                table_field._original_field_name,
                copy.deepcopy(table_field),
            )
            new_field_value: Any = fields_values.get(
                table_field._original_field_name,
                EmptyFieldValue(),
            )
            setattr(
                self,
                table_field._original_field_name,
                new_field_value,
            )

    def __getattr__(
        self: Self,
        attribute: str,
    ) -> Any:
        return self.__dict__[attribute]

    def __getattribute__(self: Self, attribute: str) -> Any:
        """Return value of the value instead of the instance of the field.

        Value of the field will be returned only
        if we have initialized `instance` of the field,
        not the class.

        ### Params
        :param attribute: the attribute we are accessing.

        ### Returns
        :returns: any attribute of the table.
        """
        table_meta = object.__getattribute__(self, "_table_meta")
        table_fields = object.__getattribute__(table_meta, "table_fields")
        if attribute in table_fields:
            return object.__getattribute__(self, attribute).value
        return super().__getattribute__(attribute)

    @classmethod
    def _parse_table_fields(
        cls: type[MetaTable],
    ) -> dict[str, Field[Any]]:
        table_fields: Final[dict[str, Field[Any]]] = {
            field_class._field_data.field_name: field_class
            for field_class in cls.__dict__.values()
            if isinstance(field_class, Field)
        }

        return table_fields

    @classmethod
    def _retrieve_not_abstract_subclasses(
        cls: type[MetaTable],
    ) -> list[type[MetaTable]]:
        return [
            subclass
            for subclass in cls._subclasses
            if not subclass._table_meta.abstract
        ]



"""All Fields."""
from qaspen.fields.complex import JsonbField, JsonField
from qaspen.fields.primitive import (
    BigIntField,
    BigSerialField,
    BooleanField,
    CharField,
    DateField,
    DecimalField,
    DoublePrecisionField,
    IntegerField,
    IntervalField,
    NumericField,
    RealField,
    SerialField,
    SmallSerialField,
    TextField,
    TimeField,
    TimestampField,
    VarCharField,
)

__all__ = [
    "BigIntField",
    "BigSerialField",
    "BooleanField",
    "CharField",
    "DateField",
    "DecimalField",
    "DoublePrecisionField",
    "IntegerField",
    "IntervalField",
    "JsonField",
    "JsonbField",
    "NumericField",
    "RealField",
    "SerialField",
    "SmallSerialField",
    "TextField",
    "TimeField",
    "TimestampField",
    "VarCharField",
]



from __future__ import annotations

import dataclasses
import functools
import operator
from typing import TYPE_CHECKING, Any, Iterable

from qaspen.clauses.order_by import OrderBy
from qaspen.querystring.querystring import OrderByQueryString, QueryString
from qaspen.statements.statement import BaseStatement

if TYPE_CHECKING:
    from typing_extensions import Self

    from qaspen.fields.base import Field


@dataclasses.dataclass
class OrderByStatement(BaseStatement):
    """OrderBy statement for high-level statements.

    It is used in Select/Update/Insert/Delete Statements.

    `order_by_expressions` contains all created order_bys.
    """

    order_by_expressions: list[OrderBy] = dataclasses.field(
        default_factory=list,
    )

    def order_by(
        self: Self,
        field: Field[Any] | None = None,
        ascending: bool = True,
        nulls_first: bool = True,
        order_by_statements: Iterable[OrderBy] | None = None,
    ) -> None:
        """Create new `OrderBy`.

        ### Parameters:
        - `field`: field to order by.
        - `ascending`: `ASC` or `DESC` order.
        - `nulls_first`: `NULL` first or not.
        - `order_by_statements`: already initialized OrderBys.
        """
        if field:
            self.order_by_expressions.append(
                OrderBy(
                    field=field,
                    ascending=ascending,
                    nulls_first=nulls_first,
                ),
            )

        if order_by_statements:
            self.order_by_expressions.extend(
                order_by_statements,
            )

    def querystring(self: Self) -> QueryString:
        """Build `QueryString`."""
        if not self.order_by_expressions:
            return QueryString.empty()
        final_order_by: OrderByQueryString = functools.reduce(
            operator.add,
            [
                order_by_expression.querystring()
                for order_by_expression in self.order_by_expressions
            ],
        )

        return QueryString(
            *final_order_by.template_arguments,
            sql_template=f"ORDER BY {final_order_by.sql_template}",
        )
